# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) Contributors to the OpenEXR Project.

from subprocess import PIPE, run
import sys

def do_run(cmd, expect_error = False, data=None):
    cmd_string = " ".join(cmd)
    print(f"running {cmd_string}")
    if data:
        result = run (cmd, input=data, stdout=PIPE, stderr=PIPE)
    else:
        result = run (cmd, stdout=PIPE, stderr=PIPE, universal_newlines=True)
    if expect_error:
        if result.returncode != 0:
            return result
        print(f"error: {cmd_string} did not fail as expected")
        print(f"stdout:\n{result.stdout}")
        print(f"stderr:\n{result.stderr}")
        sys.exit(1)
            
    if result.returncode != 0:
        print(f"error: {cmd_string} failed: returncode={result.returncode}")
        print(f"stdout:\n{result.stdout}")
        print(f"stderr:\n{result.stderr}")
        sys.exit(1)
    return result



#!/usr/bin/env python3

#
# SPDX-License-Identifier: BSD-3-Clause
# Copyright Contributors to the OpenEXR Project.
#

# This is the example code from src/wrappers/python/README.md

import os
import OpenEXR
import numpy as np

def test_write():

    # Generate arrays for R, G, and B channels with random values
    height, width = (20, 10)
    R = np.random.rand(height, width).astype('f')
    G = np.random.rand(height, width).astype('f')
    B = np.random.rand(height, width).astype('f')
    channels = { "R" : R, "G" : G, "B" : B }
    header = { "compression" : OpenEXR.ZIP_COMPRESSION,
               "type" : OpenEXR.scanlineimage }

    with OpenEXR.File(header, channels) as outfile:
        outfile.write("readme.exr")

    os.remove("readme.exr")
    print("ok")

def test_write_RGB():

    # Generate a 3D NumPy array for RGB channels with random values
    height, width = (20, 10)
    RGB = np.random.rand(height, width, 3).astype('f')

    channels = { "RGB" : RGB }
    header = { "compression" : OpenEXR.ZIP_COMPRESSION,
               "type" : OpenEXR.scanlineimage }

    with OpenEXR.File(header, channels) as outfile:
        outfile.write("readme.exr")

    print("ok")

def test_read_separate_channels():

    with OpenEXR.File("readme.exr", separate_channels=True) as infile:

        header = infile.header()
        print(f"type={header['type']}")
        print(f"compression={header['compression']}")

        R = infile.channels()["R"].pixels
        G = infile.channels()["G"].pixels
        B = infile.channels()["B"].pixels
        height, width = R.shape
        for y in range(height):
            for x in range(width):
                pixel = (R[y, x], G[y, x], B[y, x])
                print(f"pixel[{y}][{x}]={pixel}")

    print("ok")

def test_read_RGB():

    with OpenEXR.File("readme.exr") as infile:

        print(f"readme.exr: {infile.channels()}")
        
        RGB = infile.channels()["RGB"].pixels
        height, width = RGB.shape[0:2]
        for y in range(height):
            for x in range(width):
                pixel = tuple(RGB[y, x])
                print(f"pixel[{y}][{x}]={pixel}")

    print("ok")

def test_modify():

    with OpenEXR.File("readme.exr") as f:
        
        f.header()["displayWindow"] = ((3,4),(5,6))
        f.header()["screenWindowCenter"] = np.array([1.0,2.0],'float32')
        f.header()["comments"] = "test image"
        f.header()["longitude"] = -122.5
        f.write("readme_modified.exr")

        with OpenEXR.File("readme_modified.exr") as o:
            dw = o.header()["displayWindow"]
            assert (tuple(dw[0]), tuple(dw[1])) == ((3,4),(5,6))
            swc = o.header()["screenWindowCenter"]
            assert tuple(swc) == (1.0, 2.0)
            assert o.header()["comments"] == "test image"
            assert o.header()["longitude"] == -122.5

    os.remove("readme.exr")
    print("ok")

def test_multipart_write():

    height, width = (20, 10)
    Z0 = np.zeros((height, width), dtype='f')
    Z1 = np.ones((height, width), dtype='f')

    P0 = OpenEXR.Part({}, {"Z" : Z0 })
    P1 = OpenEXR.Part({}, {"Z" : Z1 })

    f = OpenEXR.File([P0, P1])
    f.write("readme_2part.exr")

    with OpenEXR.File("readme_2part.exr") as o:
        assert o.parts[0].name() == "Part0"
        assert o.parts[0].width() == 10
        assert o.parts[0].height() == 20
        assert o.parts[1].name() == "Part1"
        assert o.parts[1].width() == 10
        assert o.parts[1].height() == 20

    print("ok")

def test_multipart_write():

    height, width = (20, 10)

    Z0 = np.zeros((height, width), dtype='f')
    P0 = OpenEXR.Part(header={"type" : OpenEXR.scanlineimage },
                      channels={"Z" : Z0 })

    Z1 = np.ones((height, width), dtype='f')
    P1 = OpenEXR.Part(header={"type" : OpenEXR.scanlineimage },
                      channels={"Z" : Z1 })

    f = OpenEXR.File(parts=[P0, P1])
    f.write("readme_2part.exr")

    with OpenEXR.File("readme_2part.exr") as o:
        assert o.parts[0].name() == "Part0"
        assert o.parts[0].type() == OpenEXR.scanlineimage
        assert o.parts[0].width() == 10
        assert o.parts[0].height() == 20
        assert np.array_equal(o.parts[0].channels["Z"].pixels, Z0)
        assert o.parts[1].name() == "Part1"
        assert o.parts[1].type() == OpenEXR.scanlineimage
        assert o.parts[1].width() == 10
        assert o.parts[1].height() == 20
        assert np.array_equal(o.parts[1].channels["Z"].pixels, Z1)

    print("ok")

def test_write_tiled():

    height, width = (20, 10)

    Z = np.zeros((height, width), dtype='f')
    P = OpenEXR.Part({"type" : OpenEXR.tiledimage,
                      "tiles" : OpenEXR.TileDescription() },
                     {"Z" : Z })

    with OpenEXR.File([P]) as f:
        f.write("readme_tiled.exr")

    with OpenEXR.File("readme_tiled.exr") as o:
        assert o.parts[0].name() == ""
        assert o.parts[0].type() == OpenEXR.tiledimage

def test_write_deep():
    
    height, width = (20, 10)

    Z = np.empty((height, width), dtype=object)
    for y in range(height):
        for x in range(width):
            Z[y, x] = np.array([y*width+x], dtype='uint32')

    channels = { "Z" : Z }
    header = { "compression" : OpenEXR.ZIPS_COMPRESSION,
               "type" : OpenEXR.deepscanline }
    with OpenEXR.File(header, channels) as outfile:
        outfile.write("readme_test_tiled_deep.exr")

def test_read_deep():

    with OpenEXR.File("readme_test_tiled_deep.exr") as infile:

        Z = infile.channels()["Z"].pixels
        height, width = Z.shape
        for y in range(height):
            for x in range(width):
                for z in Z[y,x]:
                    print(f"deep sample at {y},{x}: {z}")

    os.remove("readme_test_tiled_deep.exr")


if __name__ == '__main__':

    test_multipart_write()
    test_write()
    test_write_RGB()
    test_read_separate_channels()
    test_read_RGB()
    test_modify()
    test_write_tiled()
    test_write_deep()
    test_read_deep()



#!/usr/bin/env python

# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) Contributors to the OpenEXR Project.

import sys, os, tempfile, atexit
from do_run import do_run

print(f"testing exrmaketiled: {' '.join(sys.argv)}")

exrmaketiled = sys.argv[1]
exrinfo = sys.argv[2]
image_dir = sys.argv[3]
version = sys.argv[4]

test_images = {}
test_images["GammaChart"] = f"{image_dir}/GammaChart.exr"

assert(os.path.isfile(exrmaketiled)), "\nMissing " + exrmaketiled
assert(os.path.isfile(exrinfo)), "\nMissing " + exrinfo
assert(os.path.isdir(image_dir)), "\nMissing " + image_dir
assert(os.path.isfile(test_images["GammaChart"])), "\nMissing " + image

fd, outimage = tempfile.mkstemp(".exr")
os.close(fd)

def cleanup():
    print(f"deleting {outimage}")
atexit.register(cleanup)

# no args = usage message
result = do_run ([exrmaketiled], True)
assert result.stderr.startswith ("Usage: ")

# -h = usage message
result = do_run ([exrmaketiled, "-h"])
assert result.stdout.startswith ("Usage: ")

result = do_run ([exrmaketiled, "--help"])
assert result.stdout.startswith ("Usage: ")

# --version
result = do_run ([exrmaketiled, "--version"])
assert result.stdout.startswith ("exrmaketiled")
assert version in result.stdout

result = do_run ([exrmaketiled, test_images["GammaChart"], outimage])
assert os.path.isfile(outimage)

result = do_run ([exrinfo, "-v", outimage])
assert 'tiled image has levels: x 1 y 1' in result.stdout

print("success")



#!/usr/bin/env python

# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) Contributors to the OpenEXR Project.

import sys, os, tempfile, atexit
from do_run import do_run

print(f"testing exrmakepreview: {' '.join(sys.argv)}")

exrmakepreview = sys.argv[1]
exrinfo = sys.argv[2]
image_dir = sys.argv[3]
version = sys.argv[4]

test_images = {}
test_images["GrayRampsHorizontal"] = f"{image_dir}/GrayRampsHorizontal.exr"

assert(os.path.isfile(exrmakepreview)), "\nMissing " + exrmakepreview
assert(os.path.isfile(exrinfo)), "\nMissing " + exrinfo
assert(os.path.isdir(image_dir)), "\nMissing " + image_dir

# no args = usage message
result = do_run ([exrmakepreview], True)
assert result.stderr.startswith ("Usage: ")

# -h = usage message
result = do_run ([exrmakepreview, "-h"])
assert result.stdout.startswith ("Usage: ")

result = do_run ([exrmakepreview, "--help"])
assert result.stdout.startswith ("Usage: ")

# --version
result = do_run ([exrmakepreview, "--version"])
assert result.stdout.startswith ("exrmakepreview")
assert version in result.stdout

def find_line(keyword, lines):
    for line in lines:
        if line.startswith(keyword):
            return line
    return None

fd, outimage = tempfile.mkstemp(".exr")
os.close(fd)

def cleanup():
    print(f"deleting {outimage}")
atexit.register(cleanup)

result = do_run ([exrmakepreview, "-w", "50", "-e", "1", "-v", test_images["GrayRampsHorizontal"], outimage])

result = do_run ([exrinfo, "-v", outimage])
output = result.stdout.split('\n')
assert "preview 50 x 50" in find_line("  preview", output)

print("success")












#!/usr/bin/env python

# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) Contributors to the OpenEXR Project.

import sys, os, tempfile, atexit, json
from do_run import do_run

print(f"testing exrmetrics: {' '.join(sys.argv)}")

exrmetrics = sys.argv[1]
image_dir = sys.argv[3]
version = sys.argv[4]

assert(os.path.isfile(exrmetrics)), "\nMissing " + exrmetrics
assert(os.path.isdir(image_dir)), "\nMissing " + image_dir

test_images = {}
test_images["GrayRampsHorizontal"] = f"{image_dir}/GrayRampsHorizontal.exr"
test_images["multipart"] = f"{image_dir}/multipart.0001.exr"
test_images["Flowers"] = f"{image_dir}/Flowers.exr"

fd, outimage = tempfile.mkstemp(".exr")
os.close(fd)

def cleanup():
    print(f"deleting {outimage}")
atexit.register(cleanup)

# no args = usage message
result = do_run ([exrmetrics], True)
assert result.stderr.startswith ("Usage: ")

# -h = usage message
result = do_run  ([exrmetrics, "-h"])
assert result.stdout.startswith ("Usage: ")

result = do_run  ([exrmetrics, "--help"])
assert result.stdout.startswith ("Usage: ")

# --version
result = do_run  ([exrmetrics, "--version"])
assert result.stdout.startswith ("exrmetrics")
assert version in result.stdout

# test missing arguments, using just the -option but no value

for a in ["-p","-l","-16","-z","-t","-i","--passes","-o","--pixelmode","--time"]:
    result = do_run  ([exrmetrics, a], True)
    assert "Missing" in result.stderr

for image in test_images:
    for time in ["none","read","write","reread","read,write","read,reread","read,write,reread"]:
        for passes in ["1","2"]:
            for nosize in range(0,2):
              command = [exrmetrics]
              command += ["-i",test_images[image], "--passes",passes,"--time",time,"-o",outimage]
              if nosize:
                  command += ['--no-size']
              result = do_run (command)
              assert os.path.isfile(outimage)
              if len(result.stdout):
                # confirm data is valid JSON (will not be true if filename contains quotes)
                data = json.loads(result.stdout)
                assert len(data) == 1
                if not nosize:
                  for x in ['file','pixels','compression','part type','total raw size']:
                     assert x in data[0]

# test -p 0 with multipart file
result = do_run ([exrmetrics, test_images["multipart"], "-p", "0", "-z","zip"])
data = json.loads(result.stdout)
assert(len(data)==1),"\n Unexpected list size in JSON object"

print("success")



#!/usr/bin/env python

# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) Contributors to the OpenEXR Project.

import sys, os
from do_run import do_run

print(f"testing exrinfo: {' '.join(sys.argv)}")

exrinfo = sys.argv[1]
image_dir = sys.argv[2]
version = sys.argv[3]

test_images = {}
test_images["GrayRampsHorizontal"] = f"{image_dir}/GrayRampsHorizontal.exr"

result = do_run ([exrinfo, "-h"])
assert result.stdout.startswith ("Usage: ")

result = do_run ([exrinfo, "--help"])
assert result.stdout.startswith ("Usage: ")

# --version
result = do_run ([exrinfo, "--version"])
assert result.stdout.startswith ("exrinfo")
assert version in result.stdout

result = do_run ([exrinfo, test_images["GrayRampsHorizontal"], "-a", "-v"])
output = result.stdout.split('\n')
try:
    assert ('pxr24' in output[1])
    assert ('800 x 800' in output[2])
    assert ('800 x 800' in output[3])
    assert ('1 channels' in output[4])
except AssertionError:
    print(result.stdout)
    raise

# test image as stdio
with open(test_images["GrayRampsHorizontal"], 'rb') as f:
    image_data = f.read()
result = do_run ([exrinfo, '-', "-a", "-v"], data=image_data)
output = result.stdout.decode().split('\n')
try:
    assert ('pxr24' in output[1])
    assert ('800 x 800' in output[2])
    assert ('800 x 800' in output[3])
    assert ('1 channels' in output[4])
except AssertionError:
    print(result.stdout)
    raise

print("success")




#!/usr/bin/env python

# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) Contributors to the OpenEXR Project.

import sys, os
from do_run import do_run

print(f"testing exrcheck: {' '.join(sys.argv)}")

exrcheck = sys.argv[1]
image_files = sys.argv[2:]

if not os.path.isfile(exrcheck) or not os.access(exrcheck, os.X_OK):
    print(f"error: no such file: {exrcheck}")
    sys.exit(1)

for exr_file in image_files:

    do_run([exrcheck, exr_file])
    do_run([exrcheck, "-m", exr_file])
    do_run([exrcheck, "-t", exr_file])
    do_run([exrcheck, "-s", exr_file])
    do_run([exrcheck, "-c", exr_file])

print("success.")
sys.exit(0)




#!/usr/bin/env python

# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) Contributors to the OpenEXR Project.

import sys, os, tempfile, atexit
from do_run import do_run

print(f"testing exrmultipart: {' '.join(sys.argv)}")

exrmultipart = sys.argv[1]
exrinfo = sys.argv[2]
image_dir = sys.argv[3]
version = sys.argv[4]

test_images = {}
test_images["multipart"] = f"{image_dir}/multipart.0001.exr"
test_images["singlepart"] = f"{image_dir}/GammaChart.exr"

result = do_run  ([exrmultipart], True)
assert "Usage:" in result.stderr

# -h = usage message
result = do_run  ([exrmultipart, "-h"])
assert result.stdout.startswith ("Usage: ")

result = do_run  ([exrmultipart, "--help"])
assert result.stdout.startswith ("Usage: ")

# --version
result = do_run  ([exrmultipart, "--version"])
assert result.stdout.startswith ("exrmultipart")
assert version in result.stdout

fd, outimage = tempfile.mkstemp(".exr")
os.close(fd)

def cleanup():
    print(f"deleting {outimage}")
    os.unlink(outimage)
atexit.register(cleanup)

# combine
result = do_run ([exrmultipart, "-combine", "-i", f"{test_images['multipart']}:0", f"{test_images['multipart']}:1", "-o", outimage])

result = do_run ([exrinfo, outimage])

# error: can't convert multipart images
result = do_run ([exrmultipart, "-convert", "-i", test_images["multipart"], "-o", outimage], True)

# convert
result = do_run ([exrmultipart, "-convert", "-i", test_images["singlepart"], "-o", outimage])

result = do_run  ([exrinfo, outimage])

# separate

# get part names from the multipart image
result = do_run  ([exrinfo, test_images["multipart"]])
part_names = {}
for p in result.stdout.split('\n part ')[1:]:
    output = p.split('\n')
    part_number, part_name = output[0].split(': ')
    part_names[part_number] = part_name

with tempfile.TemporaryDirectory() as tempdir:

    result = do_run ([exrmultipart, "-separate", "-i", test_images["multipart"], "-o", f"{tempdir}/separate"])

    for i in range(1, 10):
        s = f"{tempdir}/separate.{i}.exr"
        result = do_run ([exrinfo, "-v", s])
        output = result.stdout.split('\n')
        assert output[1].startswith(' parts: 1')
        output[2].startswith(' part 1:')
        part_name = output[2][9:]
        part_number = str(i)
        assert part_names[part_number] == part_name

print("success")




#!/usr/bin/env python3

#
# SPDX-License-Identifier: BSD-3-Clause
# Copyright Contributors to the OpenEXR Project.
#

from __future__ import print_function
import sys
import os
import random
from array import array

import OpenEXR
import Imath

test_dir = os.path.dirname(__file__)

FLOAT = Imath.PixelType(Imath.PixelType.FLOAT)
UINT = Imath.PixelType(Imath.PixelType.UINT)
HALF = Imath.PixelType(Imath.PixelType.HALF)

testList = []

#
# Write a simple exr file, read it back and confirm the data is the same.
#

def test_write_read():

    width = 100
    height = 100
    size = width * height

    h = OpenEXR.Header(width,height)
    h['b2i'] = Imath.Box2i(Imath.V2i(0, 0), Imath.V2i(10, 10))
    h['v2f'] = Imath.V2f(0.0, 1.0)
    h['li'] = Imath.LineOrder(Imath.LineOrder.RANDOM_Y)
    h['chromaticities'] = Imath.Chromaticities(Imath.V2f(0.0,1.0),Imath.V2i(2.0,3.0),Imath.V2f(4.0,5.0),Imath.V2f(6.0,7.0))
    h['channels'] = {'R' : Imath.Channel(FLOAT),
                     'G' : Imath.Channel(FLOAT),
                     'B' : Imath.Channel(FLOAT),
                     'A' : Imath.Channel(FLOAT)} 
    h['tc'] = Imath.TimeCode(0,1,2,3)
    h['kc'] = Imath.KeyCode(1,2,3,4,5)
    h['r'] = Imath.Rational(1,3)
    h['c'] =  Imath.Compression(Imath.Compression.ZIP_COMPRESSION)
    h['td'] = Imath.TileDescription(64,64)
    o = OpenEXR.OutputFile(f"{test_dir}/write.exr", h)
    r = array('f', [n for n in range(size*0,size*1)]).tobytes()
    g = array('f', [n for n in range(size*1,size*2)]).tobytes()
    b = array('f', [n for n in range(size*2,size*3)]).tobytes()
    a = array('f', [n for n in range(size*3,size*4)]).tobytes()
    channels = {'R' : r, 'G' : g, 'B' : b, 'A' : a}
    o.writePixels(channels)
    o.close()

    i = OpenEXR.InputFile(f"{test_dir}/write.exr")
    ih = i.header()
    assert r == i.channel('R')
    assert g == i.channel('G')
    assert b == i.channel('B')
    assert a == i.channel('A')
    assert ih['b2i'] == h['b2i']
    assert ih['v2f'] == h['v2f']
    assert ih['chromaticities'].red == h['chromaticities'].red
    assert ih['chromaticities'].green == h['chromaticities'].green
    assert ih['chromaticities'].blue == h['chromaticities'].blue
    assert ih['chromaticities'].white == h['chromaticities'].white
    assert ih['tc'] == h['tc']
    assert ih['kc'] == h['kc']
    assert ih['r'] == h['r']
    assert ih['c'] == h['c']
    assert ih['td'] == h['td']

testList.append(("test_write_read", test_write_read))

def test_level_modes():

    assert Imath.LevelMode("ONE_LEVEL").v == Imath.LevelMode(Imath.LevelMode.ONE_LEVEL).v
    assert Imath.LevelMode("MIPMAP_LEVELS").v == Imath.LevelMode(Imath.LevelMode.MIPMAP_LEVELS).v
    assert Imath.LevelMode("RIPMAP_LEVELS").v == Imath.LevelMode(Imath.LevelMode.RIPMAP_LEVELS).v

testList.append(("test_level_modes", test_level_modes))

#
# Write an image as UINT, read as FLOAT, and the reverse.
#
def test_conversion():
    codemap = { 'f': FLOAT, 'I': UINT }
    original = [0, 1, 33, 79218]
    for frm_code,to_code in [ ('f','I'), ('I','f') ]:
        hdr = OpenEXR.Header(len(original), 1)
        hdr['channels'] = {'L': Imath.Channel(codemap[frm_code])}
        x = OpenEXR.OutputFile(f"{test_dir}/out.exr", hdr)
        x.writePixels({'L': array(frm_code, original).tobytes()})
        x.close()

        xin = OpenEXR.InputFile(f"{test_dir}/out.exr")
        assert array(to_code, xin.channel('L', codemap[to_code])).tolist() == original
    os.remove(f"{test_dir}/out.exr")

testList.append(("test_conversion", test_conversion))

#
# Confirm failure on reading from non-exist location
#

def test_invalid_input():
    try:
        OpenEXR.InputFile("/bad/place")
    except:
        pass
    else:
        assert 0

testList.append(("test_invalid_input", test_invalid_input))

#
# Confirm failure on writing to invalid location
#

def test_invalid_output():

    try:
        hdr = OpenEXR.Header(640, 480)
        OpenEXR.OutputFile("/bad/place", hdr)
    except:
        pass
    else:
        assert 0

testList.append(("test_invalid_output", test_invalid_output))

def test_one():
    oexr = OpenEXR.InputFile(f"{test_dir}/write.exr")

    header = oexr.header()

    default_size = len(oexr.channel('R'))
    half_size = len(oexr.channel('R', Imath.PixelType(Imath.PixelType.HALF)))
    float_size = len(oexr.channel('R', Imath.PixelType(Imath.PixelType.FLOAT)))
    uint_size = len(oexr.channel('R', Imath.PixelType(Imath.PixelType.UINT)))

    assert default_size in [ half_size, float_size, uint_size]
    assert float_size == uint_size
    assert (float_size / 2) == half_size

    assert len(oexr.channel('R',
                            pixel_type = FLOAT,
                            scanLine1 = 10,
                            scanLine2 = 10)) == (4 * (header['dataWindow'].max.x + 1))


    data = b" " * (4 * 100 * 100)
    h = OpenEXR.Header(100,100)
    x = OpenEXR.OutputFile(f"{test_dir}/out.exr", h)
    x.writePixels({'R': data, 'G': data, 'B': data})
    x.close()
    os.remove(f"{test_dir}/out.exr")

testList.append(("test_one", test_one))

#
# Check that the channel method and channels method return the same data
#

def test_channel_channels():

    aexr = OpenEXR.InputFile(f"{test_dir}/write.exr")
    acl = sorted(aexr.header()['channels'].keys())
    a = [aexr.channel(c) for c in acl]
    b = aexr.channels(acl)

    assert a == b

testList.append(("test_channel_channels", test_channel_channels))

def test_types():
    for original in [ [0,0,0], list(range(10)), list(range(100,200,3)) ]:
        for code,t in [ ('I', UINT), ('f', FLOAT) ]:
            data = array(code, original).tobytes()
            hdr = OpenEXR.Header(len(original), 1)
            hdr['channels'] = {'L': Imath.Channel(t)}

            x = OpenEXR.OutputFile(f"{test_dir}/out.exr", hdr)
            x.writePixels({'L': data})
            x.close()

            xin = OpenEXR.InputFile(f"{test_dir}/out.exr")
            # Implicit type
            assert array(code, xin.channel('L')).tolist() == original
            # Explicit typen
            assert array(code, xin.channel('L', t)).tolist() == original
            # Explicit type as kwarg
            assert array(code, xin.channel('L', pixel_type = t)).tolist() == original
    os.remove(f"{test_dir}/out.exr")

testList.append(("test_types", test_types))

def test_invalid_pixeltype():
    oexr = OpenEXR.InputFile(f"{test_dir}/write.exr")
    FLOAT = Imath.PixelType.FLOAT
    try:
        f.channel('R',FLOAT)
    except:
        pass
    else:
        assert 0
    os.remove(f"{test_dir}/write.exr")

testList.append(("test_invalid_pixeltype", test_invalid_pixeltype))

#
# Write arbitrarily named channels.
#

def test_write_mchannels():
    hdr = OpenEXR.Header(100, 100)
    for chans in [ set("a"), set(['foo', 'bar']), set("abcdefghijklmnopqstuvwxyz") ]:
        hdr['channels'] = dict([(nm, Imath.Channel(Imath.PixelType(Imath.PixelType.FLOAT))) for nm in chans])
        x = OpenEXR.OutputFile(f"{test_dir}/out0.exr", hdr)
        data = array('f', [0] * (100 * 100)).tobytes()
        x.writePixels(dict([(nm, data) for nm in chans]))
        x.close()
        assert set(OpenEXR.InputFile(f"{test_dir}/out0.exr").header()['channels']) == chans
    os.remove(f"{test_dir}/out0.exr")

testList.append(("test_write_mchannels", test_write_mchannels))

def load_red(filename):
    oexr = OpenEXR.InputFile(filename)
    return oexr.channel('R')

#
# Write the pixels to two images, first as a single call,
# then as multiple calls.  Verify that the images are identical.
#

def test_write_chunk():
    for w,h,step in [(100, 10, 1), (64,48,6), (1, 100, 2), (640, 480, 4)]:
        data = array('f', [ random.random() for x in range(w * h) ]).tobytes()

        hdr = OpenEXR.Header(w,h)
        x = OpenEXR.OutputFile(f"{test_dir}/out0.exr", hdr)
        x.writePixels({'R': data, 'G': data, 'B': data})
        x.close()

        hdr = OpenEXR.Header(w,h)
        x = OpenEXR.OutputFile(f"{test_dir}/out1.exr", hdr)
        for y in range(0, h, step):
            subdata = data[y * w * 4:(y+step) * w * 4]
            x.writePixels({'R': subdata, 'G': subdata, 'B': subdata}, step)
        x.close()

        oexr0 = load_red(f"{test_dir}/out0.exr")
        oexr1 = load_red(f"{test_dir}/out1.exr")
        assert oexr0 == oexr1
    os.remove(f"{test_dir}/out0.exr")
    os.remove(f"{test_dir}/out1.exr")

testList.append(("test_write_chunk", test_write_chunk))

for test in testList:
    funcName = test[0]
    print(f"calling {funcName}")
    test[1]()

print("ok")



#!/usr/bin/env python

# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) Contributors to the OpenEXR Project.

import sys, os, tempfile, atexit
from do_run import do_run

print(f"testing exrstdattr: {' '.join(sys.argv)}")

exrstdattr = sys.argv[1]
exrinfo = sys.argv[2]
image_dir = sys.argv[3]
version = sys.argv[4]

assert(os.path.isfile(exrstdattr)), "\nMissing " + exrstdattr
assert(os.path.isfile(exrinfo)), "\nMissing " + exrinfo
assert(os.path.isdir(image_dir)), "\nMissing " + image_dir

test_images = {}
test_images["GrayRampsHorizontal"] = f"{image_dir}/GrayRampsHorizontal.exr"

fd, outimage = tempfile.mkstemp(".exr")
os.close(fd)

fd, outimage2 = tempfile.mkstemp(".exr")
os.close(fd)


def cleanup():
    print(f"deleting {outimage}")
atexit.register(cleanup)

# no args = usage message
result = do_run ([exrstdattr], True)
assert result.stderr.startswith ("Usage: ")

# -h = usage message
result = do_run ([exrstdattr, "-h"])
assert result.stdout.startswith ("Usage: ")

result = do_run ([exrstdattr, "--help"])
assert result.stdout.startswith ("Usage: ")

# --version
result = do_run ([exrstdattr, "--version"])
assert result.stdout.startswith ("exrstdattr")
assert version in result.stdout

attrs = [
    ["-part", "0"],
    ["-screenWindowCenter", "42", "43"],
    ["-screenWindowWidth", "4.4"],
    ["-pixelAspectRatio", "1.7"],
    ["-wrapmodes", "clamp"],
    ["-timeCode", "12345678", "34567890"],
    ["-keyCode", "1", "2", "3", "4", "5", "6", "20"],
    ["-framesPerSecond", "48", "1"],
    ["-envmap", "LATLONG"],
    ["-isoSpeed", "2.1"],
    ["-aperture", "3.2"],
    ["-expTime", "4.3"],
    ["-focus", "5.4"],
    ["-altitude", "6.5"],
    ["-latitude", "7.6"],
    ["-longitude", "8.7"],
    ["-utcOffset", "9"],
    ["-owner", "florian"],
    ["-xDensity", "10.0"],
    ["-adoptedNeutral", "1.1", "2.2"],
    ["-whiteLuminance", "17.1"],
    ["-chromaticities", "1", "2", "3", "4", "5", "6", "7", "8"],
    ["-int", "test_int", "42"],
    ["-float", "test_float", "4.2"],
    ["-string", "test_string", "forty two"],
    ["-capDate", "1999:12:31 23:59:59"],
    ["-comments", "blah blah blah"],
]

# test missing arguments, using just the -option but no value

for a in attrs:
    result = do_run ([exrstdattr, a[0]], True)

command = [exrstdattr]
for a in attrs:
    command += a 
command += [test_images["GrayRampsHorizontal"], outimage]

result = do_run (command)
assert os.path.isfile(outimage)

result = do_run ([exrinfo, "-v", outimage])
try:
    assert('adoptedNeutral: v2f [ 1.1, 2.2 ]' in result.stdout)
    assert('altitude: float 6.5' in result.stdout)
    assert('aperture: float 3.2' in result.stdout)
    assert('capDate: string \'1999:12:31 23:59:59\'' in result.stdout)
    assert('channels: chlist 1 channels' in result.stdout)
    assert('\'Y\': half samp 1 1' in result.stdout)
    assert('chromaticities: chromaticities r[1, 2] g[3, 4] b[5, 6] w[7, 8]' in result.stdout)
    assert('comments: string \'blah blah blah\'' in result.stdout)
    assert('compression: compression \'pxr24\' (0x05)' in result.stdout)
    assert('dataWindow: box2i [ 0, 0 - 799 799 ] 800 x 800' in result.stdout)
    assert('displayWindow: box2i [ 0, 0 - 799 799 ] 800 x 800' in result.stdout)
    assert('envmap: envmap latlong' in result.stdout)
    assert('expTime: float 4.3' in result.stdout)
    assert('focus: float 5.4' in result.stdout)
    assert('framesPerSecond: rational 48 / 1 (48)' in result.stdout)
    assert('isoSpeed: float 2.1' in result.stdout)
    assert('keyCode: keycode mfgc 1 film 2 prefix 3 count 4 perf_off 5 ppf 6 ppc 20' in result.stdout)
    assert('latitude: float 7.6' in result.stdout)
    assert('lineOrder: lineOrder 0 (increasing)' in result.stdout)
    assert('longitude: float 8.7' in result.stdout)
    assert('owner: string \'florian\'' in result.stdout)
    assert('pixelAspectRatio: float 1.7' in result.stdout)
    assert('screenWindowCenter: v2f [ 42, 43 ]' in result.stdout)
    assert('screenWindowWidth: float 4.4' in result.stdout)
    assert('test_float: float 4.2' in result.stdout)
    assert('test_int: int 42' in result.stdout)
    assert('test_string: string \'forty two\'' in result.stdout)
    assert('timeCode: timecode time 305419896 user 878082192' in result.stdout)
    assert('type: string \'scanlineimage\'' in result.stdout)
    assert('utcOffset: float 9' in result.stdout)
    assert('whiteLuminance: float 17.1' in result.stdout)
    assert('wrapmodes: string \'clamp\'' in result.stdout)
    assert('xDensity: float 10' in result.stdout)
except AssertionError:
    print(result.stdout)
    raise

# test for bad command
result = do_run([exrstdattr,test_images["GrayRampsHorizontal"],outimage,"extra_argument"],True)

# test for bad erase argument
result = do_run ([exrstdattr, "-erase"], True)

# test for errors trying to delete a critical attribute
result = do_run ([exrstdattr, "-erase","dataWindow",outimage,outimage2], True)

# test deleting 'comments'
result = do_run ([exrstdattr, "-erase","comments",outimage,outimage2])
assert os.path.isfile(outimage2)

result = do_run ([exrinfo, "-v", outimage2])
assert "comments" not in result.stdout

print("success")



#!/usr/bin/env python3

# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) Contributors to the OpenEXR Project.

import pytest

def test_import():
    import OpenEXR
    assert OpenEXR.__name__ == "OpenEXR"



#!/usr/bin/env python3

#
# SPDX-License-Identifier: BSD-3-Clause
# Copyright Contributors to the OpenEXR Project.
#

#
# Download images from the openexr-images repo, and for each image,
# read it as both separate channels (default) and as RGB channels,
# write it, read the written file, and confirm it's identical.
#

from __future__ import print_function
import sys
import os
import tempfile
import atexit
import unittest
import numpy as np
import math
from subprocess import PIPE, run

import OpenEXR

def equalWithRelError (x1, x2, e):
    return ((x1 - x2) if (x1 > x2) else (x2 - x1)) <= e * (x1 if (x1 > 0) else -x1)

def compare_files(lhs, rhs):

    if len(lhs.parts) != len(rhs.parts):
        raise Exception(f"#parts differs: {len(lhs.parts)} {len(rhs.parts)}")

    for Plhs, Prhs in zip(lhs.parts,rhs.parts):
        compare_parts(Plhs, Prhs)

def is_default(name, value):

    if name == "screenWindowWidth":
        return value == 1.0

    if name == "type":
        return value == OpenEXR.scanlineimage

    return True

def compare_parts(lhs, rhs):
    
    attributes = set(lhs.header.keys()).union(set(rhs.header.keys()))

    for a in attributes:
        if a in ["channels", "chunkCount"]:
            continue

        if a not in lhs.header:
            if not is_default(a, rhs.header[a]):
                raise Exception(f"attribute {a} not in lhs header")
        elif a not in rhs.header:
            if not is_default(a, lhs.header[a]):
                raise Exception(f"attribute {a} not in rhs header")
        else:
            compare_attributes(a, lhs.header[a], rhs.header[a])

    if len(lhs.channels) != len(rhs.channels):
        raise Exception(f"#channels in {lhs.name} differs: {len(lhs.channels)} {len(rhs.channels)}")

    for c in lhs.channels.keys():
        compare_channels(lhs.channels[c], rhs.channels[c])

def compare_attributes(name, lhs, rhs):

    # convert tuples to array for comparison
    
    if isinstance(lhs, tuple):
        lhs = np.array(lhs)
        
    if isinstance(rhs, tuple):
        rhs = np.array(rhs)
        
    if isinstance(lhs, np.ndarray) and isinstance(rhs, np.ndarray):
        if lhs.shape != rhs.shape:
            raise Exception(f"attribute {name}: array shapes differ: {lhs} {rhs}")
        close = np.isclose(lhs, rhs, 1e-5, equal_nan=True)
        if not np.all(close):
            raise Exception(f"attribute {name}: arrays differ: {lhs} {rhs}")
    elif isinstance(lhs, float) and isinstance(rhs, float):
        if not equalWithRelError(lhs, rhs, 1e05):
            if math.isfinite(lhs) and math.isfinite(rhs):
                raise Exception(f"attribute {name}: floats differ: {lhs} {rhs}")
    elif lhs != rhs:
        raise Exception(f"attribute {name}: values differ: {lhs} {rhs}")

def compare_channels(lhs, rhs):

    if (lhs.name != rhs.name or
        lhs.type() != rhs.type() or
        lhs.xSampling != rhs.xSampling or
        lhs.ySampling != rhs.ySampling):
        raise Exception(f"channel {lhs.name} differs: {lhs.__repr__()} {rhs.__repr__()}")
    if lhs.pixels.shape != rhs.pixels.shape:
        raise Exception(f"channel {lhs.name}: image size differs: {lhs.pixels.shape} vs. {rhs.pixels.shape}")
        
    if lhs.pixels.dtype == np.dtype('O') and rhs.pixels.dtype == np.dtype('O'):
        height, width = lhs.pixels.shape
        for y in range(height):
            for x in range(width):
                ld = lhs.pixels[y,x]
                rd = rhs.pixels[y,x]
                if ld is None and rd is None:
                    continue
                if ld.shape != rd.shape:
                    raise Exception(f"channel {lhs.name}: deep pixels {i} differ: {ld} {rd}")
                with np.errstate(invalid='ignore'):
                    close = np.isclose(ld, rd, 1e-5, equal_nan=True)
                    if not np.all(close):
                        for i in np.argwhere(close==False):
                            raise Exception(f"channel {lhs.name}: pixels[{y}][{x}] deep sample{i} differs: {ld} {rd}")
    else:
                
        with np.errstate(invalid='ignore'):
            close = np.isclose(lhs.pixels, rhs.pixels, 1e-5, equal_nan=True)

        if not np.all(close):
            for i in np.argwhere(close==False):
                y,x = i
                lp = lhs.pixels[y,x]
                rp = rhs.pixels[y,x]
                if math.isfinite(lp) and math.isfinite(rp):
                    raise Exception(f"channel {lhs.name}: pixels {i} differ: {lp} {rp}")

exr_files = [
    "TestImages/GammaChart.exr",
    "TestImages/SquaresSwirls.exr",
    "TestImages/GrayRampsDiagonal.exr",
    "TestImages/BrightRingsNanInf.exr",
    "TestImages/WideFloatRange.exr",
    "TestImages/GrayRampsHorizontal.exr",
    "TestImages/WideColorGamut.exr",
    "TestImages/BrightRings.exr",
    "TestImages/RgbRampsDiagonal.exr",
    "TestImages/AllHalfValues.exr",
    "Beachball/multipart.0007.exr",
    "Beachball/singlepart.0007.exr",
    "Beachball/singlepart.0006.exr",
    "Beachball/multipart.0006.exr",
    "Beachball/multipart.0004.exr",
    "Beachball/singlepart.0004.exr",
    "Beachball/singlepart.0005.exr",
    "Beachball/multipart.0005.exr",
    "Beachball/singlepart.0001.exr",
    "Beachball/multipart.0001.exr",
    "Beachball/singlepart.0002.exr",
    "Beachball/multipart.0002.exr",
    "Beachball/multipart.0003.exr",
    "Beachball/singlepart.0003.exr",
    "Beachball/multipart.0008.exr",
    "Beachball/singlepart.0008.exr",
    "DisplayWindow/t12.exr",
    "DisplayWindow/t06.exr",
    "DisplayWindow/t07.exr",
    "DisplayWindow/t13.exr",
    "DisplayWindow/t05.exr",
    "DisplayWindow/t11.exr",
    "DisplayWindow/t10.exr",
    "DisplayWindow/t04.exr",
    "DisplayWindow/t14.exr",
    "DisplayWindow/t15.exr",
    "DisplayWindow/t01.exr",
    "DisplayWindow/t03.exr",
    "DisplayWindow/t02.exr",
    "DisplayWindow/t16.exr",
    "DisplayWindow/t09.exr",
    "DisplayWindow/t08.exr",
    "Tiles/GoldenGate.exr",
    "Tiles/Spirals.exr",
    "Tiles/Ocean.exr",
    "v2/Stereo/composited.exr",
    "v2/Stereo/Trunks.exr",
    "v2/Stereo/Balls.exr",
    "v2/Stereo/Ground.exr",
    "v2/Stereo/Leaves.exr",
    "v2/LeftView/Trunks.exr",
    "v2/LeftView/Balls.exr",
    "v2/LeftView/Ground.exr",
    "v2/LeftView/Leaves.exr",
    "v2/LowResLeftView/composited.exr",
    "v2/LowResLeftView/Trunks.exr",
    "v2/LowResLeftView/Balls.exr",
    "v2/LowResLeftView/Ground.exr",
    "v2/LowResLeftView/Leaves.exr",
    "MultiResolution/Kapaa.exr",
    "MultiResolution/KernerEnvCube.exr",
    "MultiResolution/WavyLinesLatLong.exr",
    "MultiResolution/PeriodicPattern.exr",
    "MultiResolution/ColorCodedLevels.exr",
    "MultiResolution/MirrorPattern.exr",
    "MultiResolution/Bonita.exr",
    "MultiResolution/OrientationLatLong.exr",
    "MultiResolution/StageEnvLatLong.exr",
    "MultiResolution/WavyLinesCube.exr",
    "MultiResolution/StageEnvCube.exr",
    "MultiResolution/KernerEnvLatLong.exr",
    "MultiResolution/WavyLinesSphere.exr",
    "MultiResolution/OrientationCube.exr",
    "Chromaticities/Rec709_YC.exr",
    "Chromaticities/XYZ_YC.exr",
    "Chromaticities/XYZ.exr",
    "Chromaticities/Rec709.exr",
    "ScanLines/Desk.exr",
    "ScanLines/Blobbies.exr",
    "ScanLines/CandleGlass.exr",
    "ScanLines/PrismsLenses.exr",
    "ScanLines/Tree.exr",
    "ScanLines/Cannon.exr",
    "ScanLines/MtTamWest.exr",
    "ScanLines/StillLife.exr",
    "MultiView/Fog.exr",
    "MultiView/Adjuster.exr",
    "MultiView/Balls.exr",
    "MultiView/Impact.exr",
    "MultiView/LosPadres.exr",
]

#
# These don't work yet, so skip them.
# 

bug_files = [
    "LuminanceChroma/MtTamNorth.exr", # channel BY differs.
    "Chromaticities/Rec709_YC.exr",  # channel BY differs.
    "Chromaticities/XYZ_YC.exr",  # channel BY differs.
]

bug_files = []
exr_files = [
    "v2/Stereo/Trunks.exr",
]

class TestImages(unittest.TestCase):

    def download_file(self, url, output_file):

        try:
            result = run(['curl', '-o', output_file, url], stdout=PIPE, stderr=PIPE, universal_newlines=True)
            print(" ".join(result.args))
            if result.returncode != 0:
                print(result.stderr)
                return False
        except Exception as e:
            print(f"Download of {url} failed: {e}")
            return False
        return True

    def print_file(self, f, print_pixels = False):

        print(f"file {f.filename}")
        print(f"parts:")
        parts = f.parts
        for p in parts:
            print(f"  part: {p.name()} {p.type()} {p.compression()} height={p.height()} width={p.width()}")
            h = p.header
            for a in h:
                print(f"    header[{a}] {h[a]}")
            for n,c in p.channels.items():
                print(f"    channel[{c.name}] shape={c.pixels.shape} strides={c.pixels.strides} {c.pixels.dtype}")
                if print_pixels:
                    maxy, maxx = c.pixels.shape[0], c.pixels.shape[1]
                    maxy, maxx = 2, 800
                    for y in range(maxy):
                        for x in range(maxx):
                            print(f"{n}[{y},{x}]={c.pixels[y,x]}")
                        
    def print_channel_names(self, file):
        for p in file.parts:
            s = f"part[{p.part_index}] name='{p.name()}', channels: ["
            for c in p.channels:
                s += f" {c}"
            s += " ]"
            print(s)

    def do_test_tiled(self, url):

        #
        # Write the image as tiled, reread and confirm it's the same
        #

        if "://" not in url:
            filename = url
        else:
            filename = "test_file.exr"
            if not self.download_file(url, filename):
                return

        print(f"Reading {url} ...")
        f = OpenEXR.File(filename)

        # Set the type and tile description (default)
        for P in f.parts:
            if P.header["type"] in [OpenEXR.deepscanline,  OpenEXR.deeptiled]:
                return
            
            P.header["compression"] = OpenEXR.ZIP_COMPRESSION
            P.header["type"] = OpenEXR.tiledimage
            if "tiles" not in P.header:
                P.header["tiles"] = OpenEXR.TileDescription()
                for n,C in P.channels.items():
                    C.xSampling = 1
                    C.ySampling = 1

        f.write("tiled.exr")

        t = OpenEXR.File("tiled.exr")

        # Clear the chunkCount values before comparison, since they'll
        # differ on conversion from scanline to tiled.
        for P in f.parts and t.parts:
            if "chunkCount" in P.header: del P.header["chunkCount"]
        for P in t.parts:
            if "chunkCount" in P.header: del P.header["chunkCount"]

        print(f"Comparing original to tiled...")
        compare_files(f, t)

    def do_test_image(self, url):

        verbose = False
        verbose_pixels = False

        print(f"testing image {url}...")
        
        if "://" not in url:
            filename = url
        else:
            filename = "test_file.exr"
            if not self.download_file(url, filename):
                return

        # Read the file as separate channels, as usual...

        print(f"Reading {url} as separate channels...")
        with OpenEXR.File(filename) as separate_channels:
            if verbose:
                print("separate_channels:")
                self.print_file(separate_channels, verbose_pixels)
                self.print_channel_names(separate_channels)

            # Write it out

            print(f"Writing separate_channels.exr...")
            separate_channels.write("separate_channels.exr")

            # Read the file that was just written
            print(f"Reading {url} as separate channels...")
            with OpenEXR.File("separate_channels.exr") as separate_channels2:
                if verbose:
                    self.print_channel_names(separate_channels2)

                # Confirm that the file that was just written is identical to the original

                print(f"Comparing separate_channels to separate_channels2...")
                compare_files(separate_channels, separate_channels2)
                print(f"Comparing separate_channels to separate_channels2...done.")

                # Read the original file as RGBA channels

                print(f"Reading {url} as rgba channels...")
                with OpenEXR.File(filename, True) as rgba_channels:
                    if verbose:
                        self.print_channel_names(rgba_channels)

                    # Write it out

                    print(f"Writing rgba_channels.exr...")
                    rgba_channels.write("rgba_channels.exr")

                # Read the file that was just written (was RGBA in memory,
                # should have been written as usual)

                print(f"Reading rgba_channels as separate channels3...")
                with OpenEXR.File("rgba_channels.exr") as separate_channels3:
                    if verbose:
                        print("separate_channels3:")
                        self.print_file(separate_channels3, verbose_pixels)
                        self.print_channel_names(separate_channels3)

                    # Confirm that it, too, is the same as the original

                    print(f"Comparing separate_channels to separate_channels3...")
                    compare_files(separate_channels, separate_channels3)

        print("ok.")

    def test_images(self):

        #
        # Run the test only if the OPENEXR_TEST_IMAGE_REPO env var is set
        #

        REPO_VAR = "OPENEXR_TEST_IMAGE_REPO" 
        if REPO_VAR in os.environ:

            REPO = os.environ[REPO_VAR]

            for filename in exr_files:

                if filename in bug_files:
                    print(f"skipping bug file: {filename}")
                    continue

                url = f"{REPO}/{filename}"

                self.do_test_image(url)
#                self.do_test_tiled(url)

            print("OK")

        else:    

            print(f"{sys.argv[0]}: skipping images, no repo")

if __name__ == '__main__':


    if len(sys.argv) > 2:
        exr_files = sys.argv[2:]
        bug_files = []
        
    unittest.main()
    print("OK")





#!/usr/bin/env python

# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) Contributors to the OpenEXR Project.

import sys, os, tempfile, atexit
from do_run import do_run

print(f"testing exr2aces: {' '.join(sys.argv)}")

exr2aces = sys.argv[1]
exrinfo = sys.argv[2]
image_dir = sys.argv[3]
version = sys.argv[4]

test_images = {}
test_images["GrayRampsHorizontal"] = f"{image_dir}/GrayRampsHorizontal.exr"

if not os.path.isfile(exr2aces) or not os.access(exr2aces, os.X_OK):
    print(f"error: no such file: {exr2aces}")
    sys.exit(1)

# no args = usage message, error
result = do_run ([exr2aces], True)
assert result.stderr.startswith ("Usage: ")

# -h = usage message
result = do_run ([exr2aces, "-h"])
assert result.stdout.startswith ("Usage: ")

result = do_run ([exr2aces, "--help"])
assert result.stdout.startswith ("Usage: ")

# --version
result = do_run ([exr2aces, "--version"])
assert result.stdout.startswith ("exr2aces")
assert version in result.stdout

# invalid arguments
result = do_run ([exr2aces, "foo.exr", "bar.exr"], True)

def find_line(keyword, lines):
    for line in lines:
        if line.startswith(keyword):
            return line
    return None

fd, outimage = tempfile.mkstemp(".exr")
os.close(fd)

def cleanup():
    print(f"deleting {outimage}")
atexit.register(cleanup)

result = do_run ([exr2aces, "-v", test_images["GrayRampsHorizontal"], outimage])

result = do_run ([exrinfo, "-v", outimage])

# confirm the output has the proper chromaticities
assert "chromaticities: chromaticities r[0.7347, 0.2653] g[0, 1] b[0.0001, -0.077] w[0.32168, 0.33767]" in result.stdout

print("success")












#!/usr/bin/env python

# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) Contributors to the OpenEXR Project.

import sys, os, tempfile, atexit
from do_run import do_run

print(f"testing exrenvmap: {' '.join(sys.argv)}")

exrenvmap = sys.argv[1]
exrinfo = sys.argv[2]
image_dir = sys.argv[3]
version = sys.argv[4]

test_images = {}
test_images["latlong"] = f"{image_dir}/WavyLinesLatLong.exr"
test_images["cube"] = f"{image_dir}/WavyLinesCube.exr"

assert(os.path.isfile(exrenvmap)), "\nMissing " + exrenvmap
assert(os.path.isfile(exrinfo)), "\nMissing " + exrinfo
assert(os.path.isdir(image_dir)), "\nMissing " + image_dir
assert(os.path.isfile(test_images["latlong"])), "\nMissing " + test_images["latlong"]

fd, outimage = tempfile.mkstemp(".exr")
os.close(fd)

def cleanup():
    print(f"deleting {outimage}")
atexit.register(cleanup)

# no args = usage message
result = do_run ([exrenvmap], True)
assert result.stderr.startswith ("Usage: ")

# -h = usage message
result = do_run ([exrenvmap, "-h"])
assert result.stdout.startswith ("Usage: ")

# --help = usage message
result = do_run ([exrenvmap, "--help"])
assert result.stdout.startswith ("Usage: ")

# --version
result = do_run ([exrenvmap, "--version"])
assert result.stdout.startswith ("exrenvmap")
assert version in result.stdout

# default
result = do_run ([exrenvmap, test_images["latlong"], outimage])
assert os.path.isfile(outimage)
default_file_size = os.path.getsize(outimage)

result = do_run ([exrenvmap, "-li", test_images["latlong"], outimage])
assert os.path.isfile(outimage)

result = do_run ([exrenvmap, "-ci", test_images["cube"], outimage])
assert os.path.isfile(outimage)

# -o 
result = do_run ([exrenvmap, "-o", test_images["latlong"], outimage])
assert os.path.isfile(outimage)

result = do_run ([exrinfo, "-v", outimage])
assert 'tiles: tiledesc size 64 x 64 level 0 (single image) round 0 (down)' in result.stdout
os.unlink(outimage)

# -m 
result = do_run ([exrenvmap, "-m", test_images["latlong"], outimage])
assert os.path.isfile(outimage)

result = do_run ([exrinfo, "-v", outimage])
assert 'tiles: tiledesc size 64 x 64 level 1 (mipmap) round 0 (down)' in result.stdout
os.unlink(outimage)

# -c 
result = do_run ([exrenvmap, "-c", test_images["latlong"], outimage])
assert os.path.isfile(outimage)

result = do_run ([exrinfo, "-v", outimage])
assert 'envmap: envmap cube' in result.stdout
os.unlink(outimage)

# -l 
result = do_run ([exrenvmap, "-l", test_images["latlong"], outimage])
assert os.path.isfile(outimage)

result = do_run ([exrinfo, "-v", outimage])
assert 'envmap: envmap latlong' in result.stdout
os.unlink(outimage)

# -w 

result = do_run ([exrenvmap, "-w", test_images["latlong"], outimage], True)
assert not os.path.isfile(outimage)

result = do_run ([exrenvmap, "-w", "-64", test_images["latlong"], outimage], True)
assert not os.path.isfile(outimage)

result = do_run ([exrenvmap, "-w", "64", test_images["latlong"], outimage])
assert os.path.isfile(outimage)

result = do_run ([exrinfo, "-v", outimage])
assert 'x tile count: 1 (sz 64)' in result.stdout
assert 'y tile count: 6 (sz 384)' in result.stdout
os.unlink(outimage)

# -f (filter)
result = do_run ([exrenvmap, "-f", test_images["latlong"], outimage], True)
assert not os.path.isfile(outimage)

result = do_run ([exrenvmap, "-f", "1.1", test_images["latlong"], outimage], True)
assert not os.path.isfile(outimage)

result = do_run ([exrenvmap, "-f", "-1.1", "6", test_images["latlong"], outimage], True)
assert not os.path.isfile(outimage)

result = do_run ([exrenvmap, "-f", "1.1", "-6", test_images["latlong"], outimage], True)
assert not os.path.isfile(outimage)

result = do_run ([exrenvmap, "-f", "1.1", "6", test_images["latlong"], outimage])
assert os.path.isfile(outimage)
file_size = os.path.getsize(outimage)
assert file_size != default_file_size
os.unlink(outimage)

# -b (blur)
result = do_run ([exrenvmap, "-b", test_images["latlong"], outimage])
assert os.path.isfile(outimage)
file_size = os.path.getsize(outimage)
assert file_size != default_file_size
os.unlink(outimage)

# -t 
result = do_run ([exrenvmap, "-t", test_images["latlong"], outimage], True)
assert not os.path.isfile(outimage)

result = do_run ([exrenvmap, "-t", "32", test_images["latlong"], outimage], True)
assert not os.path.isfile(outimage)

result = do_run ([exrenvmap, "-t", "-32", "48", test_images["latlong"], outimage], True)
assert not os.path.isfile(outimage)

result = do_run ([exrenvmap, "-t", "32", "-48", test_images["latlong"], outimage], True)
assert not os.path.isfile(outimage)

result = do_run ([exrenvmap, "-t", "32", "48", test_images["latlong"], outimage])
assert os.path.isfile(outimage)

result = do_run ([exrinfo, "-v", outimage])
assert 'x tile count: 8 (sz 256)' in result.stdout
assert 'y tile count: 32 (sz 1536)' in result.stdout
os.unlink(outimage)

# -u 
result = do_run ([exrenvmap, "-u", test_images["latlong"], outimage])
assert os.path.isfile(outimage)

result = do_run ([exrinfo, "-v", outimage])
assert 'tiles: tiledesc size 64 x 64 level 0 (single image) round 1 (up)' in result.stdout
os.unlink(outimage)

# -z 
result = do_run ([exrenvmap, "-z", test_images["latlong"], outimage], True)
assert not os.path.isfile(outimage)

result = do_run ([exrenvmap, "-z", "xxx", test_images["latlong"], outimage], True)
assert not os.path.isfile(outimage)

for z in ["none", "rle", "zip", "piz", "pxr24", "b44", "b44a", "dwaa", "dwab"]:
    result = do_run ([exrenvmap, "-z", z, test_images["latlong"], outimage])
    assert os.path.isfile(outimage)

    result = do_run ([exrinfo, "-v", outimage])
    assert f'compression: compression \'{z}\'' in result.stdout
    os.unlink(outimage)

with tempfile.TemporaryDirectory() as tempdir:

    cube_face_image_t = f"{tempdir}/out.%.exr"
    result = do_run ([exrenvmap, test_images["latlong"], cube_face_image_t])

    for o in ["+X", "-X", "+Y", "-Y", "+Z", "-Z"]:
        
        cube_face_image = f"{tempdir}/out.{o}.exr"
        result = do_run ([exrinfo, "-v", cube_face_image])
        assert 'x tile count: 4 (sz 256)' in result.stdout
        assert 'y tile count: 4 (sz 256)' in result.stdout

    result = do_run ([exrenvmap, cube_face_image_t, outimage])

    result = do_run ([exrinfo, "-v", outimage])
    assert 'tiles: tiledesc size 64 x 64 level 0 (single image) round 0 (down)' in result.stdout

print("success")



#!/usr/bin/env python3

#
# SPDX-License-Identifier: BSD-3-Clause
# Copyright Contributors to the OpenEXR Project.
#

from __future__ import print_function
import sys
import os
import tempfile
import atexit
import unittest
import numpy as np

import OpenEXR

def print_deep(outfile):
    for n,c in outfile.channels().items():
        for y in range(c.pixels.shape[0]):
            for x in range(c.pixels.shape[1]):
                d = c.pixels[y,x]
                print(f"{n}[{y},{x}]: {d}")

def compare_files(lhs, rhs):

    for Plhs, Prhs in zip(lhs.parts,rhs.parts):
        compare_parts(Plhs, Prhs)

def compare_parts(lhs, rhs):

    if len(lhs.channels) != len(rhs.channels):
        raise Exception(f"#channels in {lhs.name()} differs: {len(lhs.channels)} {len(rhs.channels)}")

    for c in lhs.channels.keys():
        compare_channels(lhs.channels[c], rhs.channels[c])

def compare_channels(lhs, rhs):

    if (lhs.name != rhs.name or
        lhs.type() != rhs.type() or
        lhs.xSampling != rhs.xSampling or
        lhs.ySampling != rhs.ySampling):
        raise Exception(f"channel {lhs.name} differs: {lhs.__repr__()} {rhs.__repr__()}")

    compare_channel_pixels(lhs, rhs)
    
def compare_channel_pixels(lhs, rhs):

    if lhs.pixels.shape != rhs.pixels.shape:
        raise Exception(f"channel {lhs.name}: image size differs: {lhs.pixels.shape} vs. {rhs.pixels.shape}")
        
    height = lhs.pixels.shape[0]
    width = lhs.pixels.shape[1]
    for y in range(height):
        for x in range(width):
            l = lhs.pixels[y,x]
            r = rhs.pixels[y,x]
            if l is None and r is None:
                continue
            close = np.isclose(l, r, 1e-5)
            if not np.all(close):
                for i in np.argwhere(close==False):
                    y,x = i
                    if math.isfinite(lhs.pixels[y,x]) and math.isfinite(rhs.pixels[y,x]):
                        raise Exception(f"channel {lhs.name}: deep pixels {i} differ: {lhs.pixels[y,x]} {rhs.pixels[y,x]}")

class TestDeep(unittest.TestCase):

    def test_deep_rgba(self):

        self.do_test_deep_rgba(False, 'uint32')
        self.do_test_deep_rgba(True, 'uint32')
        self.do_test_deep_rgba(False, 'e')
        self.do_test_deep_rgba(True, 'e')
        self.do_test_deep_rgba(False, 'float32')
        self.do_test_deep_rgba(True, 'float32')
        
    def do_test_deep_rgba(self, do_alpha, dt):

        dataWindow = ((100,100), (120,130))
        height = dataWindow[1][1] - dataWindow[0][1] + 1
        width = dataWindow[1][0] - dataWindow[0][0] + 1
        
        R = np.empty((height, width), dtype=object)
        G = np.empty((height, width), dtype=object)
        B = np.empty((height, width), dtype=object)
        if do_alpha:
            A = np.empty((height, width), dtype=object)
        Z = np.empty((height, width), dtype=object)

        for y in range(height):
            for x in range(width):
                i = y*width+x + 2
                l = i % 2 
                if l == 0:
                    R[y, x] = np.array([j for j in range(1,i)], dtype=dt)
                    G[y, x] = np.array([j*10 for j in range(1,i)], dtype=dt)
                    B[y, x] = np.array([j*100 for j in range(1,i)], dtype=dt)
                    if do_alpha:
                        A[y, x] = np.array([j*100 for j in range(1,i)], dtype=dt)
                    Z[y, x] = np.array([j*2 for j in range(1,i)], dtype=dt)
                else:
                    R[y, x] = None
                    G[y, x] = None
                    B[y, x] = None
                    if do_alpha:
                        A[y, x] = None
                    Z[y, x] = None
        
        channels = { "B" : B, "G" : G, "R" : R, "Z" : Z }
        if do_alpha:
            channels["A"] = A
        header = { "compression" : OpenEXR.ZIPS_COMPRESSION,
                   "type" : OpenEXR.deepscanline,
                   "dataWindow" : dataWindow}

        filename = "write_deep.exr"
        with OpenEXR.File(header, channels) as outfile:

            outfile.write(filename)

            with OpenEXR.File(filename) as infile:

                channel = "RGBA" if do_alpha else "RGB"
                
                C = infile.channels()[channel]
                for y in range(height):
                    for x in range(width):
                        p = C.pixels[y,x]
                        if p is None:
                            assert R[y,x] is None
                        else:
                            for i in range(p.shape[0]):
                                if do_alpha:
                                    self.assertEqual(p[i,:].tolist(), [R[y,x][i], G[y,x][i], B[y,x][i], A[y,x][i]])
                                else:
                                    self.assertEqual(p[i,:].tolist(), [R[y,x][i], G[y,x][i], B[y,x][i]])
                
                infile.write("rgb_deep.exr")

            with OpenEXR.File("rgb_deep.exr", separate_channels=True) as infile:

                compare_files(infile, outfile)

        os.remove(filename)
        os.remove("rgb_deep.exr")
                
    def test_deep(self):

        dataWindow = ((100,100), (120,130))
        height = dataWindow[1][1] - dataWindow[0][1] + 1
        width = dataWindow[1][0] - dataWindow[0][0] + 1
        
        U = np.empty((height, width), dtype=object)
        H = np.empty((height, width), dtype=object)
        F = np.empty((height, width), dtype=object)
        for y in range(height):
            for x in range(width):
                i = y*width+x
                l = i % 3 
                if l == 0:
                    U[y, x] = np.array([i], dtype='uint32')
                    H[y, x] = np.array([i], dtype='float16')
                    F[y, x] = np.array([i], dtype='float32')
                else:
                    U[y, x] = None
                    H[y, x] = None
                    F[y, x] = None
        
        channels = { "U" : U, "H" : H, "F" : F }
        header = { "compression" : OpenEXR.ZIPS_COMPRESSION,
                   "type" : OpenEXR.deepscanline,
                   "dataWindow" : dataWindow}

        filename = "write_deep.exr"
        with OpenEXR.File(header, channels) as outfile:

            outfile.write(filename)

            with OpenEXR.File(filename) as infile:

                compare_files(infile, outfile)

        os.remove(filename)

    def test_tiled_deep(self):

        dataWindow = ((100,100), (120,130))
        dataWindow = ((100,100), (103,104))
        height = dataWindow[1][1] - dataWindow[0][1] + 1
        width = dataWindow[1][0] - dataWindow[0][0] + 1
        
        U = np.empty((height, width), dtype=object)
        H = np.empty((height, width), dtype=object)
        F = np.empty((height, width), dtype=object)
        for y in range(height):
            for x in range(width):
                i = y*width+x + 2
                l = i % 3 
                if l == 0:
                    U[y, x] = np.array([j for j in range(1,i)], dtype='uint32')
                    H[y, x] = np.array([j*10 for j in range(1,i)], dtype='float16')
                    F[y, x] = np.array([j*100 for j in range(1,i)], dtype='float32')
                else:
                    U[y, x] = None
                    H[y, x] = None
                    F[y, x] = None
        
        channels = { "U" : U, "H" : H, "F" : F }
        header = { "compression" : OpenEXR.ZIPS_COMPRESSION,
                   "type" : OpenEXR.deeptile,
                   "tiles" : OpenEXR.TileDescription(),
                   "dataWindow" : dataWindow}

        filename = "test_tiled_deep.exr"
        with OpenEXR.File(header, channels) as outfile:

            outfile.write(filename)

            with OpenEXR.File(filename, False) as infile:

                compare_files(infile, outfile)

        os.remove(filename)

if __name__ == '__main__':
    unittest.main()
    print("OK")





#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright Contributors to the OpenEXR Project.

# Helper script that automates parts of the release proces:
#
# `release.py notes <tag>` - extract release notes from CHANGES.md for the tagged release, print to stdout
# `release.py news <tag>` - edit the website/news.rst file to add reference to the tagged release
# `release.py draft <tag>` - create a draft release on GitHub
# `release.py candidate <tag>` - format a message about the upcoming release, print to stdout
# `release.py cherry <label>` - list merged PRs with the given label, print cherry-pick lines (oldest merge first)
# `release.py changes <tag> [pr#> ... ]` - add section to CHANGES.md with given PRs
# `release.py log` - print a `git log` annotated with the labels from each commit's associated PR, if there is one.
#
# The file `CHANGES.md` is assumed to old the release notes. When
# preparing a release, edit this file by hand to add a description of
# the new release, with a heading `## Version ...`.
#
# The file `website/news.rst` is the source for the website's news
# page. The title of the most recent news item goes in
# `website/latest_news_title.rst`, which is also included on the main
# website landing page. The text for the news blurb is bracketed by
# `.. _LatestNewsStart:` and `.. _LatestNewsEnd:`, which allows the
# main website landing page to reproduce the text.
#

import json
import sys
import re
import subprocess
from pathlib import Path
from subprocess import PIPE, run
from datetime import datetime, timedelta
import markdown

def extract_section(content, version_tag):
    """Extract the section of release notes starting with a heading
    like "Version <tag>".  Include only the text in the section
    itself, omitting any subsections, which is typically the merged
    PRs, etc..
    """

    # Regular expression to match the version title
    version_header = re.compile(rf'^##.*Version {re.escape(version_tag)}\b.*\(([^)]+)\)', re.IGNORECASE)

    # Regular expression to match subheaders (### or higher)
    subsection_header_pattern = re.compile(r'^##\s+')

    lines = content.splitlines()
    capture = False
    section_lines = []
    release_date = None

    for line in lines:

        # Look for the start of the requested version section
        m = version_header.match(line)
        if m:
            release_date = datetime.strptime(m.group(1), "%B %d, %Y")
            capture = True
            continue

        # Stop capturing when the next subsection (##) starts
        if capture and (subsection_header_pattern.match(line) or
                        "Merged Pull Requests" in line):
            break
        # Capture lines if inside the correct section and before any subsections
        if capture:
            section_lines.append(line)

    return release_date, '\n'.join(section_lines).strip()

def markdown_to_rst(markdown_text):
    """
    Convert simple Markdown text to reStructuredText (reST) manually.
    """
    # Replace markdown links [text](url) with reST format `text <url>`_
    markdown_text = re.sub(r'\[(.*?)\]\((.*?)\)', r'`\1 <\2>`_', markdown_text)

    # Convert the special symbols
    markdown_text = re.sub(r':bug:', "🐛", markdown_text, flags=re.DOTALL)
    markdown_text = re.sub(r':rocket:', "🚀", markdown_text, flags=re.DOTALL)
    markdown_text = re.sub(r':hammer_and_wrench:', "🛠️", markdown_text, flags=re.DOTALL)
    markdown_text = re.sub(r':wrench:', "🔧", markdown_text, flags=re.DOTALL)
    markdown_text = re.sub(r':sparkles:', "✨", markdown_text, flags=re.DOTALL)
    markdown_text = re.sub(r':snake:', "🐍", markdown_text, flags=re.DOTALL)
    markdown_text = re.sub(r':package:', "📦", markdown_text, flags=re.DOTALL)
    markdown_text = re.sub(r':warning:', "⚠️", markdown_text, flags=re.DOTALL)

    # Split into lines for further processing
    lines = markdown_text.splitlines()
    rst_lines = []
    for line in lines:
        # Replace bullet points (- becomes *)
        if line.startswith("- "):
            line = line.replace("- ", "* ", 1)

        # Handle other elements (bold, italics, etc.)
        line = re.sub(r'\*\*(.*?)\*\*', r'**\1**', line)  # Bold
        line = re.sub(r'\*(.*?)\*', r'*\1*', line)  # Italics

        rst_lines.append(line)

    return '\n'.join(rst_lines).strip()

def create_draft_release(tag, release_notes):
    """
    Create a draft release using the GitHub CLI.
    """

    release_tag = tag.split("-rc")[0]
    result = run(['gh', 'release', 'create', tag, '--draft', '--title', release_tag, '-F', '-'],
                 input=release_notes,
                 text=True,
                 check=True
                 )

def get_repo_url():
    try:
        # Return the upstream if this is in a fork
        result = run(['git', 'config', '--get', 'remote.upstream.url'],
                     stdout=PIPE, stderr=PIPE, universal_newlines=True)
        url = result.stdout.strip().rstrip(".git")
        if url:
            return url
        
        result = run(['git', 'config', '--get', 'remote.origin.url'],
                     stdout=PIPE, stderr=PIPE, universal_newlines=True)
        url = result.stdout.strip().rstrip(".git")
        return url
    except subprocess.CalledProcessError:
        return None  # Not a git repo or no origin remote


def cmd_cherry(label):
    """
    List merged PRs with the given GitHub label, ordered by merge time
    (oldest first). Print one line per merge commit suitable for cherry-picking.
    """
    result = run(
        [
            "gh",
            "pr",
            "list",
            "--label",
            label,
            "--state",
            "merged",
            "--limit",
            "500",
            "--json",
            "mergedAt,mergeCommit,title,number",
        ],
        stdout=PIPE,
        stderr=PIPE,
        universal_newlines=True,
        check=False,
    )
    if result.returncode != 0:
        sys.stderr.write(result.stderr or "gh pr list failed\n")
        sys.exit(1)

    try:
        prs = json.loads(result.stdout or "[]")
    except json.JSONDecodeError as e:
        print(f"Invalid JSON from gh: {e}", file=sys.stderr)
        sys.exit(1)

    merged = []
    for pr in prs:
        id = pr.get("number") or "?"
        mc = pr.get("mergeCommit") or {}
        oid = mc.get("oid")
        merged_at = pr.get("mergedAt")
        title = pr.get("title") or ""
        if not oid or not merged_at:
            continue
        merged.append((merged_at, oid, title, id))

    merged.sort(key=lambda x: x[0])

    changes_prs = ""
    for _merged_at, oid, title, id in merged:
        abbrev = oid[:7]
        title_one_line = " ".join(title.split())
        print(f"git cherry-pick {abbrev} # {title_one_line} (#{id})")
        changes_prs += f" {id}" 

    print(f"share/util/release.py changes {label} {changes_prs}")

def prev_patch_version(base_tag):
    """
    Previous release in the same major.minor line, e.g. 3.4.5 -> 3.4.4.
    Returns None if the patch segment is missing, non-numeric, or already 0.
    """
    parts = base_tag.split(".")
    if not parts or not parts[-1].isdigit():
        return None
    patch = int(parts[-1])
    if patch <= 0:
        return None
    parts[-1] = str(patch - 1)
    return ".".join(parts)


def gh_pr_view(pr_number):
    result = run(
        [
            "gh",
            "pr",
            "view",
            str(pr_number),
            "--json",
            "title,author",
        ],
        stdout=PIPE,
        stderr=PIPE,
        universal_newlines=True,
        check=False,
    )
    if result.returncode != 0:
        sys.stderr.write(result.stderr or "gh pr view failed\n")
        sys.exit(1)
    return json.loads(result.stdout)


MERGED_PR_HEADING_RE = re.compile(
    r"^###\s+Merged Pull Requests\s*:?\s*$", re.IGNORECASE
)
MERGED_WORKFLOW_HEADING_RE = re.compile(
    r"^###\s+Merged Workflow Pull Requests\s*:?\s*$", re.IGNORECASE
)
PR_BULLET_RE = re.compile(r"^\*\s*\[(\d+)\]\(")


def parse_pr_blocks_dict(lines, i, stop_at_workflow_heading):
    """
    From index *i*, read ``* [pr](url)`` entries. Each value is the bullet line and
    the following line joined with ``\\n`` when that following line is not another
    bullet or (if applicable) the workflow subsection heading.

    If *stop_at_workflow_heading* is true, stop before ``### Merged Workflow Pull Requests``.
    Returns ``(dict pr -> str, next_index)``.
    """
    out = {}
    n = len(lines)
    while i < n:
        line = lines[i]
        if stop_at_workflow_heading and MERGED_WORKFLOW_HEADING_RE.match(line.strip()):
            break
        mo = PR_BULLET_RE.match(line.strip())
        if mo:
            pr = int(mo.group(1))
            bullet = line
            if i + 1 < n:
                nxt = lines[i + 1]
                if stop_at_workflow_heading and MERGED_WORKFLOW_HEADING_RE.match(
                    nxt.strip()
                ):
                    out[pr] = bullet
                    i += 1
                    continue
                if PR_BULLET_RE.match(nxt.strip()):
                    out[pr] = bullet
                    i += 1
                    continue
                out[pr] = "\n".join([bullet, nxt])
                i += 2
                continue
            out[pr] = bullet
            i += 1
            continue
        i += 1
    return out, i


def parse_section(lines):
    """
    Parse a version slice of CHANGES.md (list of line strings, no trailing newlines).

    Returns ``heading`` (lines before ``### Merged Pull Requests``), and two dicts
    mapping PR number -> two-line bullet+title string for each subsection.
    """
    heading = []
    i = 0
    n = len(lines)

    while i < n:
        line = lines[i]
        if MERGED_PR_HEADING_RE.match(line.strip()):
            break
        heading.append(line)
        i += 1

    merged_prs = {}
    if i < n and MERGED_PR_HEADING_RE.match(lines[i].strip()):
        i += 1
        merged_prs, i = _parse_pr_blocks_dict(lines, i, stop_at_workflow_heading=True)

    merged_workflow_prs = {}
    if i < n and MERGED_WORKFLOW_HEADING_RE.match(lines[i].strip()):
        i += 1
        merged_workflow_prs, i = _parse_pr_blocks_dict(lines, i, stop_at_workflow_heading=False)

    return heading, merged_prs, merged_workflow_prs

def pr_is_workflow_only(pr_number: int) -> bool:
    """Return True if every file changed by the PR is under .github/workflows/."""
    result = run(
        ["gh", "pr", "view", str(pr_number), "--json", "files",
         "--jq", "[.files[].path]"],
        stdout=PIPE, stderr=PIPE, universal_newlines=True, check=False,
    )
    if result.returncode != 0:
        sys.stderr.write(result.stderr or "gh pr view failed\n")
        sys.exit(1)
    paths = json.loads(result.stdout or "[]")
    return bool(paths) and all(p.startswith(".github/workflows/") for p in paths)
def cmd_changes(tag, prs):
    """
    Add a section to CHANGES.md for the given release if one doesn't already exist, and add the given PR 
    to the list if it's not already there.
    """

    with open("CHANGES.md", encoding="utf-8") as f:
        lines = f.read().splitlines()

    base_tag = tag.lstrip("v").split("-rc")[0] # i.e. "3.4.7"
    prev_tag = prev_patch_version(base_tag) # i.e.e "3.4.6"

    section_re = re.compile(rf"^##\s+Version\s+{re.escape(base_tag)}\b", re.IGNORECASE)
    prev_re = re.compile(rf"^##\s+Version\s+{re.escape(prev_tag)}\b", re.IGNORECASE)

    #
    # Terminology:
    #  "header" = everything from the beginning of the file up to the new section
    #  "footer" = everything afer the new section
    #
    
    # Find the line index of the existing section if there is one, and
    # to the previous release section, i.e. the release before, which
    # comes after it in the file.
    
    section_index = None
    footer_index = None # line of the following section, i.e. the remainder of the file.
    for i, line in enumerate(lines):
        if section_index is None and section_re.match(line):
            section_index = i
        if prev_re is not None and footer_index is None and prev_re.match(line):
            footer_index = i

    # header_index is the last line of content before the new section
    header_index = ( 
        section_index if section_index is not None else footer_index
    )
    if header_index is None:
        print(
            "Could not locate insertion point (no matching version section and no "
            "previous patch release heading in CHANGES.md).",
            file=sys.stderr,
        )
        sys.exit(1)

    release_date = datetime.now() + timedelta(days=3)
    date_str = release_date.strftime("%B %e, %Y")

    if section_index is not None:
        # Parse the existing section into the heading (section heading
        # and existing notes), the dict of merged PRs, and the dict of merge workflow PRs
        section_heading, merged_prs, merged_workflow_prs = parse_section(
            lines[section_index:footer_index]
        )
    else:
        # The section does not exist, so create the stub of a new one
        section_heading = [f"## Version {base_tag} ({date_str})\n"]
        merged_prs = {}
        merged_workflow_prs = {}

    # Find the index of the table of contents entry, if there is one.
    toc, prev_toc = None, None
    toc_re = re.compile(rf"^\*\s+\[Version\s+{re.escape(base_tag)}\]", re.IGNORECASE)
    prev_toc_re = re.compile(rf"^\*\s+\[Version\s+{re.escape(prev_tag)}\]", re.IGNORECASE)
    for i, line in enumerate(lines[:footer_index]):
        if toc is None and toc_re.match(line):
            toc = i
        elif prev_toc is None and prev_toc_re.match(line):
            prev_toc = i
            break

    # Format the list entry for each new PR:
    # * [number](url)
    # title
    for pr_number in prs:
        info = gh_pr_view(pr_number)
        title = info.get("title") or ""
        author = (info.get("author") or {}).get("login") or ""
        is_workflow = "dependabot" in author or pr_is_workflow_only(pr_number)
        url = get_repo_url()
        title_one_line = " ".join(title.split())
        pr_block = f"* [{pr_number}]({url}/pull/{pr_number})\n{title_one_line}"
        # Add it to the appropriate section
        if is_workflow:
            merged_workflow_prs[pr_number] = pr_block
        else:
            merged_prs[pr_number] = pr_block

    with open("CHANGES.md", "w", encoding="utf-8", newline="\n") as f:
        if toc is None:
            # No table of contents entry for this release, so add one.
            f.write("\n".join(lines[:prev_toc]) + "\n")  # everything up to the previous release entry
            # Write new entry
            base_tag_nonum = base_tag.replace(".","")
            date_str_lower = date_str.lower().replace(",","").replace(" ","-")
            f.write(f"* [Version {base_tag}](#version-{base_tag_nonum}-{date_str_lower}) {date_str}\n")
            # Write the rest of the table of contents and everything up to the new release section
            f.write("\n".join(lines[prev_toc:header_index]) + "\n")  
        else:
            # Table of contents already has an entry, so just write the header
            f.write("\n".join(lines[:header_index]) + "\n")
        # Write the release section
        f.write("\n".join(section_heading))
        f.write("\n### Merged Pull Requests\n\n")
        for pr, value in sorted(merged_prs.items(), reverse=True):
            f.write(value + "\n")
        f.write("\n### Merged Workflow Pull Requests\n\n")
        for pr, value in sorted(merged_workflow_prs.items(), reverse=True):
            f.write(value + "\n")
        # Write the rest of the file
        f.write("\n" + "\n".join(lines[footer_index:]))

def update_news_file(release_notes, tag, release_date):
    """
    Update the website/news.rst file with the release notes.
    """
    date_str = release_date.strftime("%B %e, %Y")
    new_section_title = f"{date_str} - OpenEXR {tag} Released"

    # Extract the section header for the most recent item from website/latest_news_title.rst.
    # This needs to be added as an explicit section header.
    result = run(['git', 'show', f"HEAD:website/latest_news_title.rst"],
                 stdout=PIPE, stderr=PIPE, universal_newlines=True)
    line = result.stdout.split('\n')[-1]
    old_section_title = line.split("replace:: ")[1].rstrip().lstrip('**').rstrip('**')

    # Write the news section title to the latest_news_title.rst file,
    # so it can be include on the website front page
    with open('website/latest_news_title.rst', 'w') as f:
        f.write("..\n")
        f.write("  SPDX-License-Identifier: BSD-3-Clause\n")
        f.write("  Copyright (c) Contributors to the OpenEXR Project.\n")
        f.write(f".. |latest-news-title| replace:: **{new_section_title}**")

    result = run(['git', 'show', f"HEAD:website/news.rst"],
                 stdout=PIPE, stderr=PIPE, universal_newlines=True)
    content = result.stdout
    if content == None:
        print("No news.txt at tag {tag}")

    # Split the news text at the first header line, the one containing only ='s. Discard the ='s.
    # old_news[0] is everything up to the first ====...'
    # old_news[1] is everything after the frist ====...'
    old_news = re.split(r'^=+\s*$', content, maxsplit=1, flags=re.MULTILINE)

    # Remove the _LatestNewsStart/_LatestNewsEnd from the second half
    old_news[1] = re.sub(r'\.\. _LatestNewsStart:\n', '', old_news[1], flags=re.DOTALL)
    old_news[1] = re.sub(r'\.\. _LatestNewsEnd:\n', '', old_news[1], flags=re.DOTALL)

    # Write the new website/news.rst file, with the new section
    # inserted between the halves of the existing file, with new
    # section headers and _LastsNews markers.

    with open('website/news.rst', 'w') as f:
        f.write(old_news[0])
        f.write('='*len(new_section_title)+'\n\n')
        f.write('.. _LatestNewsStart:\n\n')
        f.write(release_notes+'\n\n')
        f.write('.. _LatestNewsEnd:\n\n')
        f.write(old_section_title+'\n')
        f.write('='*len(old_section_title))
        f.write(old_news[1])

def markdown_to_html(markdown_text):
    """
    Convert markdown content to HTML.
    """

    result = run(
        ['pandoc', '-f', 'markdown', '-t', 'html'],  # pandoc command to convert markdown to html
        input=markdown_text,                         # pass markdown_text as stdin
        text=True,                                   # treat input and output as strings (not bytes)
        capture_output=True                          # capture the output
    )

    html = f"<style> body {{ line-height: 1.0; margin: 0; padding: 0; }} p, h1, h2, h3, h4, h5, h6 {{ margin: 0; padding: 0; }} </style> {result.stdout}"

    return html


def convert_to_html(release_notes):
    """
    Convert release notes to HTML format.
    """
    notes = release_notes.replace('\n', '<br>')
    html_content = f"<h2>Release Notes</h2><p>{notes}</p>"
    return html_content

def pr_labels(line):
    """line is from `git log --pretty:format='%h %s'`, so for
    squash/merge commits, it ends in (#<pr>).

    Return the release-related labels for the identified PR, i.e.
    labels of the form 'v<major>.<minor>.<patch>'
    """

    LOG_PR_SUFFIX_RE = re.compile(r"\(#(\d+)\)\s*$")
    m = LOG_PR_SUFFIX_RE.search(line)
    if not m:
        return ""
    pr_number = m.group(1)
    result = run(
        ["gh", "pr", "view", pr_number, "--json", "labels"],
        stdout=PIPE,
        stderr=PIPE,
        universal_newlines=True,
        check=False,
    )
    if result.returncode != 0:
        return ""
    try:
        data = json.loads(result.stdout or "{}")
    except json.JSONDecodeError:
        return ""
    out = []
    for lab in data.get("labels") or []:
        name = lab.get("name") or ""
        if name.startswith("v"):
            out.append(name)
    return " ".join(out)

def cmd_log():
    """
    Run ``git log`` and print commits annotated with the release labels for the associated PRs
    """

    # Output looks like:
    # v3.4.7  3978976a Bump pypa/cibuildwheel from 3.3 to 3.4 (#2287)
    #         b19cfec2 Bump github/codeql-action from 4.32.4 to 4.32.5 (#2286)
    # v3.4.7  50ea61bd Fix build failure with glibc 2.43 due to C11 threads.h conflicts (#2262)

    result = run(
        ["git", "log", "--pretty=format:%h %s"],
        stdout=PIPE,
        stderr=PIPE,
        universal_newlines=True,
        check=False,
    )
    if result.returncode != 0:
        sys.stderr.write(result.stderr or "git log failed\n")
        sys.exit(1)

    for line in result.stdout.splitlines():
        l = pr_labels(line)
        print(f"{l:7} {line}")

def main():

    if len(sys.argv) > 1:
        action = sys.argv[1]
        if action == "log":
            cmd_log()
            return

    if len(sys.argv) < 2:
        print(
            "Usage: python release.py "
            "<notes|news|draft|candidate|cherry|changes> ...\n"
            "  release.py changes <tag> <pr-number>   e.g. release.py changes v3.4.7 1234"
        )
        sys.exit(1)

    action = sys.argv[1]

    if action == "cherry":
        if len(sys.argv) < 3:
            print(
                "Usage: python release.py cherry <label>",
                file=sys.stderr,
            )
            sys.exit(1)
        cmd_cherry(sys.argv[2])
        return

    if action == "changes":
        if len(sys.argv) < 4:
            print(
                "Usage: python release.py changes <tag> <pr-number>\n"
                "Example: python release.py changes v3.4.7 1234",
                file=sys.stderr,
            )
            sys.exit(1)
        tag = sys.argv[2]
        prs = sys.argv[3:]
        cmd_changes(tag, prs)
        return

    if len(sys.argv) < 3:
        print(
            "Usage: python release.py "
            "<notes|news|draft|candidate|cherry|changes> <tag-or-label> [date]"
        )
        sys.exit(1)

    tag = sys.argv[2]

    # Strip leading 'v' and trailing '-rc<candidate>' if necessary
    base_tag = tag.lstrip('v').split('-rc')[0]
    result = run(['git', 'tag', '--list', tag], stdout=PIPE, stderr=PIPE, universal_newlines=True)
    if result.stdout == "":
        tag += "-rc"
        result = run(['git', 'tag', '--list', tag], stdout=PIPE, stderr=PIPE, universal_newlines=True)
        if result.stdout != "":
            print(f"Using {tag} instead...")
        else:
            print(f"No such tag: {tag}")
            sys.exit(1)

    # Get the content of the CHANGES.md at the specified git tag
    try:
        result = run(['git', 'show', f"{tag}:CHANGES.md"], stdout=PIPE, stderr=PIPE, universal_newlines=True)
        content = result.stdout
        if content == None or content == "":
            print(f"No news.txt at tag {tag}")
            sys.exit(1)
    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)

    # Extract the release notes
    release_date, release_notes = extract_section(content, base_tag)
    if release_date == None:
        print("No release found.")
        return

    if action == "notes":
        print(release_notes)

    elif action == "draft":
        create_draft_release(tag, release_notes)

    elif action == "news":
        rst_text = markdown_to_rst(release_notes)
        update_news_file(rst_text, base_tag, release_date)

    elif action == "candidate":
        version = base_tag
        html_notes = markdown_to_html(release_notes)
        date_string = release_date.strftime("%A, %B %e")
        print(f"OpenEXR {version} is staged for release at tag <a href=https://github.com/AcademySoftwareFoundation/openexr/releases/tag/{tag}>{tag}</a> and will be released officially {date_string} barring any issues. <br><br> {html_notes}")

if __name__ == "__main__":
    main()



# Copyright Contributors to the MaterialX Project
# SPDX-License-Identifier: Apache-2.0
# copied from: https://github.com/AcademySoftwareFoundation/MaterialX/blob/main/python/mtx_skbuild_plugin.py
# Modifications Copyright (c) Contributors to the OpenEXR Project.

"""
This is a custom scikit-build-core plugin that will
fetch the OpenEXR version from the CMake project.
"""
import os
import tempfile
import subprocess
from pathlib import Path
from typing import FrozenSet, Dict, Optional, Union, List

from scikit_build_core.file_api.query import stateless_query
from scikit_build_core.file_api.reply import load_reply_dir


def dynamic_metadata(
    fields: FrozenSet[str],
    settings: Optional[Dict[str, object]] = None,
) -> Dict[str, Union[str, Dict[str, Optional[str]]]]:
    print("openexr_skbuild_plugin: Computing OpenEXR version from CMake...")

    if fields != "version":
        msg = f"Only the 'version' field is supported: fields={fields}"
        raise ValueError(msg)

    if settings:
        msg = "No inline configuration is supported"
        raise ValueError(msg)

    if "OPENEXR_RELEASE_CANDIDATE_TAG" in os.environ:

        # e.g. "v3.1.2-rc4"
        #
        # If OPENEXR_RELEASE_CANDIDATE_TAG is set,
        # the build is for a publish to test.pypi.org. Multiple test
        # publishes may happen in the course of preparing for a
        # release, but published packages require unique
        # names/versions, so use the release candidate tag as the
        # version (minus the leading 'v'),

        rct = os.environ["OPENEXR_RELEASE_CANDIDATE_TAG"]
        version = rct[1:]

    else:
        
        current_dir = os.path.dirname(__file__)

        with tempfile.TemporaryDirectory() as tmpdir:
            # We will use CMake's file API to get the version
            # instead of parsing the CMakeLists files.

            # First generate the query folder so that CMake can generate replies.
            reply_dir = stateless_query(Path(tmpdir))

            # Run cmake (configure). CMake will generate a reply automatically.
            try:
                subprocess.run(
                    [
                        "cmake",
                        "-S",
                        current_dir + "../../../..",
                        "-B",
                        tmpdir,
                        "-DOPENEXR_BUILD_LIBS=OFF",
                    ],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    check=True,
                    text=True,
                )
            except subprocess.CalledProcessError as exc:
                print(exc.stdout)
                raise RuntimeError(
                    "Failed to configure project to get the version"
                ) from exc

            # Get the generated replies.
            index = load_reply_dir(reply_dir)

            # Get the version from the CMAKE_PROJECT_VERSION variable.
            entries = [
                entry
                for entry in index.reply.cache_v2.entries
                if entry.name == "CMAKE_PROJECT_VERSION"
            ]

            if not entries:
                raise ValueError("Could not find OpenEXR version from CMake project")

            if len(entries) > 1:
                raise ValueError("More than one entry for CMAKE_PROJECT_VERSION found...")

            version = entries[0].value

    print("openexr_skbuild_plugin: Computed version: {0}".format(version))

    return version



#!/usr/bin/env python3

# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) Contributors to the OpenEXR Project.

#
# Generate the "Test Images" page for the openexr.com website
#
# The file "website/test_images.txt" contains urls for either .exr files
# or README.txt files from
# https://github.com/AcademySoftwareFoundation/openexr-images. This
# script processes them into .rst files with a main index listing all
# images (with thumbnails and summaries) and a separate page per image
# file showing the full image and the header contents in a table.
#
# The test_images.txt file references exr files expected to exist in
# the images repo. This script downloads the .exrs and run 'exrheader'
# on them to generate the associated .rst file, which becomes the
# descriptive content on the webpage. Each exr is expected to have an
# accompanying .jpg sidecar file for display on the webpage.
#
# The README.rst files are expected to have a heading, summary text,
# and a ".. list-table::" with descriptive text for selected
# images. The main index pages gets the main summary text, with the
# per-image descriptive text in the table next to the image thumbnail.
#
# Note that the README.rst files in the openexr-images repo are never
# actually processed directly by sphinx.
#
# The generated .rst files all go in the "website/test_images"
# directory underneath the source root. 
#
# Run this script from the source root, then commit the files to git.
#
# Note: The initial version of this process ran during cmake, ran the
# conversion to .jpg explicitly, and stored the proxy .jpg's in the
# source tree for processing by Sphinx. This avoided the need for
# permanent sidecar .jpg's in the openexr-images repo, but it
# complicated the website build, could not get the website build to
# work on Windows. Now we rely on running the exr-to-jpg conversion
# manually.
#

import sys, os, tempfile
from subprocess import PIPE, run

def exr_header(exr_path):
    '''Return a dict of the attributes in the exr file's header(s)
       The index of the dict is the part number as a string, i.e. "0", "1", etc.
       The value of each entry is a dict of attribute/value pairs.

       Also return in rows the list of all attributes across all
       parts, which equates to the complete list of rows in the table,
       since not all parts will necessarily have the same attributes.
    '''
    
    result = run (['exrheader', exr_path],
                  stdout=PIPE, stderr=PIPE, universal_newlines=True)
    if result.returncode != 0:
        raise Exception(f'failed to read header for {exr_path}')

    lines = result.stdout.split('\n') 
        
    name,version = lines[3].split(':')

    header = {}
    current_part = "0"
    current_attr = None
    
    for line in lines[4:]:
        if line.startswith(' part '):
            current_part = line[6:].rstrip(':')
            continue
        
        continuation = line.startswith("    ")
        line = line.strip()
        if len(line) == 0: # blank line
            continue
        if continuation:
            words = line.split(',')
            if len(words) > 1:
                line = words[0]
            line = line.strip().strip('"').replace('  ',' ') 
            if header[current_part][current_attr]:
                header[current_part][current_attr] += ', '
            header[current_part][current_attr] += line
            continue
        
        name_value = line.split(':')
        if len(name_value) > 1:
            current_attr = name_value[0].split(' ')[0]
            if current_part not in header:
                header[current_part] = {}
            header[current_part][current_attr] = name_value[1].strip().strip('"').replace('  ',' ')
                
    rows = set()
    for part_name,part in header.items():
        for attr_name,value in part.items():
            rows.add(attr_name)

    return header, rows

def write_rst_list_table_row(outfile, attr_name, header, rows, ignore_attr_name=False):
    '''Write a row in the rst list-table of parts/attributes on the website page for an exr file'''
    
    if ignore_attr_name:
        outfile.write(f'   * -\n')
    else:
        outfile.write(f'   * - {attr_name}\n')
    for part_number,part in header.items():
        if attr_name in part:
            value = part[attr_name]
            if type(value) is list:
                value = ' '.join(value)
        else:
            value = ''
        outfile.write(f'     - {value}\n')

def write_exr_page(rst_lpath, exr_url, exr_filename, exr_lpath, jpg_url, readme):
    '''Write the website page for each exr: title, image, and list-table of attribute name/value for each part
    
       rst_lpath:    the name of .rst file to write, including directory relative to the source root
       exr_url:      the url of exr
       exr_filename: the exr filename without directory
       exr_lpath:    the exr filename with directory relative to the source root
       jpg_url:      the url of the jpg image (from https://raw.githubusercontent.com/AcademySoftwareFoundation/openexr-images)

       Return a dict of of the header parts/attributes
    '''
    
    # Download the exr to a temp file
    fd, local_exr = tempfile.mkstemp(".exr")
    os.close(fd)

    header = None
    
    try:
        
        # Download the exr via wget
        
        print(f'wget {exr_url}')
        result = run (['wget', exr_url, '-O', local_exr], 
                      stdout=PIPE, stderr=PIPE, universal_newlines=True)
        if result.returncode != 0 or not os.path.isfile(local_exr):
            raise Exception(f'failed to read {exr_url}: no such file {local_exr}')
        
        # Read the header
        
        header, rows = exr_header(local_exr)

        os.remove(local_exr)

        if not header:
            raise Exception(f'can\'t read header for {local_exr}')
        
        print(f'write {rst_lpath}')

        with open(rst_lpath, 'w') as rst_file:

            # copyright
            rst_file.write(f'..\n')  
            rst_file.write(f'  SPDX-License-Identifier: BSD-3-Clause\n')
            rst_file.write(f'  Copyright Contributors to the OpenEXR Project.\n')
            rst_file.write(f'\n')

            # :download:
            rst_file.write(f'{exr_filename}\n')
            for i in range(0,len(exr_filename)):
                rst_file.write('#')
            rst_file.write(f'\n')
            rst_file.write(f'\n')
            rst_file.write(f':download:`{exr_url}<{exr_url}>`\n')
            rst_file.write(f'\n')

            # .. image::
            rst_file.write(f'.. image:: {jpg_url}\n')
            rst_file.write(f'   :target: {exr_url}\n')
            rst_file.write(f'\n')

            if readme:
                notes = readme_notes(readme, exr_filename, False)
                if notes:
                    rst_file.write(f'\n')
                    rst_file.write(notes)
                    rst_file.write(f'\n')
            
            # .. list-table::
            rst_file.write(f'.. list-table::\n')
            rst_file.write(f'   :align: left\n')
        
            # For multi-part files, add a header. For single-part, put the 'name' attribute at the top, if there is one
            if 'name' in rows:
                if len(header) > 2:
                    rst_file.write(f'   :header-rows: 1\n')
                    rst_file.write(f'\n')
                    write_rst_list_table_row(rst_file, 'name', header, rows, True)
                else:
                    rst_file.write(f'\n')
                    write_rst_list_table_row(rst_file, 'name', header, rows)
            else:
                rst_file.write(f'\n')

            # Write each attribute
            for attr_name in rows:
                if attr_name == 'name':
                    continue
                write_rst_list_table_row(rst_file, attr_name, header, rows)

    except Exception as e:

        os.remove(local_exr)

        raise e
    
    return header

def write_readme(index_file, repo, tag, lpath):
    '''Download the README.txt file and write its contents up to the first ".. list-table::"
       Return a list of all lines in the file.
    '''

    # Download to a temp file
    
    fd, local_readme = tempfile.mkstemp(".rst")

    try:
        
        # Download via wget
        
        readme_url = f'{repo}/{tag}/{lpath}' 
        result = run (['wget', readme_url, '-O', local_readme], 
                      stdout=PIPE, stderr=PIPE, universal_newlines=True)
        if result.returncode != 0:
            raise FileNotFoundError(result.stderr)
    
        text = ''
        found = False
        with open(local_readme, 'r') as readme_file:
            lines = readme_file.readlines()
            for line in lines:
                if '.. list-table::' in line:
                    break
                index_file.write(line[:-1])
                index_file.write('\n')

            os.unlink(local_readme)
            return lines

    except Exception as e:

        os.unlink(local_readme)
        
        raise e
    
    return None
        
def readme_notes(readme, exr_filename, html):
    '''Extract the section of the README.rst file's list-table for the given exr file.

       The list-table line lists the exr as a row header, i.e. with "* -" prefix. Return the lines up to the next row.

       Return None if there's no entry.
    '''

    found = False
    text = ''
    
    for line in readme:
        if line.startswith('   * -'):
            if exr_filename in line:
                found = True
            elif found:
                break
        else:
            if found and line != '\n':
                if not line.startswith(' '):
                    break
                if html:
                    text += '             '
                    text += line[7:].replace(' ``',' <b>').replace('``','</b>')
                else:
                    text += line[7:]
    return text

def write_exr_to_index(index_file, repo, tag, exr_lpath, readme):
    '''Write an entry to the index.rst file for the give exr, in raw html format.

       index_file: the open index.rst file descriptor
       repo:       the repo url
       tag:        the tag/branch of the repo
       exr_lpath:  the name of the exr file, with directory relative to the repo root
       readme:     the lines of the README.txt, returned by write_readme()
    '''
    
    # Examples:
    #    repo = 'https://raw.githubusercontent.com/AcademySoftwareFoundation/openexr-images'
    #    tag = 'main'
    #    exr_lpath = v2/LeftView/Ground.exr

    test_images = 'website/test_images/'
    output_dirname = test_images + os.path.dirname(exr_lpath)    # website/test_images/v2/LeftView
    os.makedirs(output_dirname, exist_ok=True)
    base_path = os.path.splitext(exr_lpath)[0]                   # v2/LeftView/Ground
    exr_filename = os.path.basename(exr_lpath)                   # Ground.exr
    exr_basename = os.path.splitext(exr_filename)[0]             # Ground
    exr_dirname = os.path.dirname(exr_lpath)                     # v2/LeftView
    rst_lpath = f'{test_images}{exr_dirname}/{exr_basename}.rst' # website/test_images/v2/LeftView/Ground.rst
    jpg_url = f'{repo}/{tag}/{base_path}.jpg'
    exr_url = f'{repo}/{tag}/{exr_lpath}'
    
    # Write the exr page
    
    header = write_exr_page(rst_lpath, exr_url, exr_filename, exr_lpath, jpg_url, readme)

    if not header:
        raise Exception(f'no header for {exr_lpath}')
    
    num_parts = len(header)
    num_channels = 0
    for p,v in header.items():
        num_channels += len(v)

    # open the row
    index_file.write('     <tr>\n')

    # row for the image
    index_file.write(f'          <td style="vertical-align: top; width:250px">\n')
    index_file.write(f'              <a href={base_path}.html> <img width="250" src="{jpg_url}"> </a>\n') 
    index_file.write(f'          </td>\n')

    # row for the summary
    index_file.write(f'          <td style="vertical-align: top; width:250px">\n')
    index_file.write(f'            <b> {exr_filename} </b>\n')
    index_file.write(f'            <ul>\n')
    if num_parts == 1:
        index_file.write(f'                <li> single part </li>\n')
    else:
        index_file.write(f'                <li> {num_parts} parts </li>\n')
    if num_parts == 1:
        index_file.write(f'                <li> 1 channel </li>\n')
    else:
        index_file.write(f'                <li> {num_channels} channels </li>\n')

    if "type" in header["0"]:
        index_file.write(f'                <li> {header["0"]["type"]} </li>\n')
    if "compression" in header["0"]:
        compression = header["0"]["compression"]
        if compression == "zip, individual scanlines":
            compression = "zip"
        elif compression == "zip, multi-scanline blocks":
            compression = "zips"
        index_file.write(f'                <li> {compression} compression </li>\n')
    if "envmap" in header["0"]:
        index_file.write(f'                <li> {header["0"]["envmap"]} </li>\n')

    index_file.write(f'            </ul>\n')
    index_file.write(f'          </td>\n')

    # row for the readme notes, if there are any
    
    if readme:
        notes = readme_notes(readme, exr_filename, True)
        if notes:
            index_file.write(f'          <td style="vertical-align: top; width:400px">\n')
            index_file.write(f'          <p>\n')
            index_file.write(notes)
            index_file.write(f'          </p>\n')
            index_file.write(f'          </td>\n')

    # close the row
    index_file.write('     </tr>\n')

    return base_path

def write_table_open(index_file):

    index_file.write(f'\n')
    index_file.write(f'.. raw:: html\n')
    index_file.write(f'\n')
    index_file.write(f'   <embed>\n')
    index_file.write(f'   <table>\n')
    index_file.write(f'\n')
    
def write_table_close(index_file):

    index_file.write(f'\n')
    index_file.write(f'   </table>\n')
    index_file.write(f'   </embed>\n')
    index_file.write(f'\n')
    
print(f'generating rst for test images ...')

print(f'PATH={os.environ["PATH"]}')
result = run (['which', 'exrheader'],
              stdout=PIPE, stderr=PIPE, universal_newlines=True)
print(f'exrheader={result.stdout}')

repo = sys.argv[1] if len(sys.argv) > 1 else 'https://raw.githubusercontent.com/AcademySoftwareFoundation/openexr-images'
tag = sys.argv[2] if len(sys.argv) > 2 else 'main'

try:
    
    with open('website/test_images/index.rst', 'w') as index_file:

        index_file.write('Test Images\n')
        index_file.write('###########\n')
        index_file.write('\n')
        index_file.write('.. toctree::\n')
        index_file.write('   :caption: Test Images\n')
        index_file.write('   :maxdepth: 2\n')
        index_file.write('\n')
        index_file.write('   toctree\n')
        index_file.write('\n')

        toctree = []
        readme = None
        table_opened = False
    
        # Process each url in the .txt file
        
        with open('website/test_images.txt', 'r') as test_images_file:
            for line in test_images_file.readlines():

                if line.startswith('#'):
                    continue
                
                lpath = line.strip('\n')
            
                if os.path.basename(lpath) == "README.rst":
                    
                    if table_opened:
                        write_table_close(index_file)
                        table_opened = False

                    readme = write_readme(index_file, repo, tag, lpath)
                
                elif lpath.endswith('.exr'):

                    if not table_opened:
                        write_table_open(index_file)
                        table_opened = True
                    
                    base_path = write_exr_to_index(index_file, repo, tag, lpath, readme)
                    if base_path:
                        toctree.append(base_path)

            if table_opened:
                write_table_close(index_file)

        # Write the toctree file, one entry per .exr page
        
        with open('website/test_images/toctree.rst', 'w') as toctree_file:
            toctree_file.write('..\n')
            toctree_file.write('  SPDX-License-Identifier: BSD-3-Clause\n')
            toctree_file.write('  Copyright Contributors to the OpenEXR Project.\n')
            toctree_file.write('\n')
            toctree_file.write('.. toctree::\n')
            toctree_file.write('   :maxdepth: 0\n')
            toctree_file.write('   :hidden:\n')
            toctree_file.write('\n')
            for t in toctree:
                toctree_file.write(f'   {t}\n')

except Exception as e:

    print(f'error: {str(e)}', file=sys.stderr)
    exit(-1)

exit(0)

                




#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright Contributors to the OpenColorIO Project.

#
# This script compares an install_manifest.txt from a CI build and
# compares it against a known valid manifest to verify that the proper
# files have been installed. Both missing files and extra files not
# expected to be installed are reported as errors.
#
# This script is invoked by the .github/workflows/ci_steps.yml workflow.
#
# The first argument is the candidate manifest from the build in
# question.  The second argument is the reference manifest from
# share/ci/install_manifest, known to be correct.
#
# The reference manifest may contain variables for the library version:
#   $MAJOR = OPENEXR_VERSION_MAJOR
#   $MINOR = OPENEXR_VERSION_MINOR
#   $PATCH = OPENEXR_VERSION_PATCH
#   $SOVERSION = OPENEXR_LIB_SOVERSION
# The values of these variables are extracted from the raw source
# files where they are maintained. This is so the reference manifests
# don't have be updated on every release, sinced these values appear
# in the installed library names.
#
# The manifests *must* be updated whenever files are added to what is
# installed: new headers, libraries, tools, cmake, etc, or a change to
# file names. Note that this validates just the file *names*, not the
# contents.
#
# The CI uploads the install manifests as job artifacts, attached to
# the GitHub status page.
#
# NOTE: If you add a new installed header, you must add it to each of
# the reference manifest in share/ci/install_manifest.
#
# The first line of the manifest file is the cmake command for the
# build that generated it, just to help identify the contents.
#

import os
import sys
import re
import argparse
from pathlib import Path

def openexr_soversion():

    with open("CMakeLists.txt", "r") as f:
        for line in f:
            m = re.search(r"set\(OPENEXR_LIB_SOVERSION (\d+)\)", line)
            if m:
                return m.group(1)

    print("Can't find soversion")
    sys.exit(1)

def openexr_version():

    major = ""
    minor = ""
    patch = ""

    with open("src/lib/OpenEXRCore/openexr_version.h", "r") as f:
        for line in f:
            pattern = r"#\s*define\s+OPENEXR_VERSION_MAJOR\s+(\d+)"
            m = re.search(pattern, line)
            if m:
                major = m.group(1)
                continue
            pattern = r"#\s*define\s+OPENEXR_VERSION_MINOR\s+(\d+)"
            m = re.search(pattern, line)
            if m:
                minor = m.group(1)
                continue
            pattern = r"#\s*define\s+OPENEXR_VERSION_PATCH\s+(\d+)"
            m = re.search(pattern, line)
            if m:
                patch = m.group(1)
                continue

    if not major or not minor or not patch:
        print("Can't find openexr version!")
        sys.exit(1)

    return major, minor, patch

def process_line(line):
    """
    Replaces all occurrences of $VAR or ${VAR} with the value of VAR
    if VAR is set in the environment.

    Args:
        line (str): The input string containing potential environment variables.

    Returns:
        str: The string with environment variables replaced by their values.
    """
    # The regular expression looks for two patterns:
    # 1. \$([a-zA-Z0-9]+): dollar sign followed by alpha-numeric characters (no underscore allowed).
    # 2. \${(\w+)}: dollar sign, a literal opening brace, one or more word characters, iteral closing brace. 
    # The | operator combines these two patterns.
    pattern = r'\$(?:([a-zA-Z0-9]+)|{(\w+)})'

    def replacer(match):
        var_name = match.group(1)
        if var_name:
            var_val = os.environ.get(var_name, match.group(0))
            if var_val == None:
                print(f"no such env var: {var_name}")
                sys.exit(1)
            return var_val
        var_name = match.group(2)
        if var_name:
            var_val = os.environ.get(var_name, match.group(0))
            if var_val == None:
                print(f"no such env var: {var_name}")
                sys.exit(1)
            return var_val
        return match.group(0)
    return re.sub(pattern, replacer, line).rstrip()

def load_manifest(path):
    """Load and return the list of files from the install manifest."""
    with open(path, 'r') as file:
        return sorted(process_line(line) for line in file if line[0]!='#')

def validate_install(candidate_manifest_path, reference_manifest_path, options):
    """Main function to verify the installed files."""

    candidate_manifest = load_manifest(candidate_manifest_path)
    reference_manifest = load_manifest(reference_manifest_path)

    print("reference_manifest:")
    for l in reference_manifest:
        print(f"  {l}")
    print("candidate_manifest:")
    for l in candidate_manifest:
        print(f"  {l}")

    # Compare manifests
    missing_files = sorted(set(reference_manifest) - set(candidate_manifest))
    extra_files = sorted(set(candidate_manifest) - set(reference_manifest))

    # Output results
    if missing_files:
        print("Error: The following files should have been installed but weren't:\n  " + '\n  '.join(missing_files))
    if extra_files:
        print("Error: The following files were installed but were not expected:\n  " + '\n  '.join(extra_files))

    if missing_files or extra_files:
        return 1

    print("install manifest is valid.")

    return 0

if __name__ == "__main__":

    print(f"validate_install: {sys.argv}")

    parser = argparse.ArgumentParser(description="Validate installed files against reference install manifest.")
    parser.add_argument("candidate_manifest", help="Path to the candidate install_manifest.txt")
    parser.add_argument("reference_manifest", help="Path to the reference install_manifest.txt")

    args = parser.parse_args()

    if not os.path.exists(args.candidate_manifest):
        print(f"candidate manifest does not exist: {args.candidate_manifest}")
        sys.exit(1)

    if not os.path.exists(args.reference_manifest):
        print(f"reference manifest does not exist: {args.reference_manifest}")
        sys.exit(1)

    print(f"candidate_manifest={args.candidate_manifest}")
    print(f"reference_manifest={args.reference_manifest}")

    status = validate_install(args.candidate_manifest, args.reference_manifest, args)

    sys.exit(status)




#!/usr/bin/env python3

#
# SPDX-License-Identifier: BSD-3-Clause
# Copyright Contributors to the OpenEXR Project.
#

from __future__ import print_function
import sys
import os
import numpy as np
import unittest

import OpenEXR

test_dir = os.path.dirname(__file__)

class TestExceptions(unittest.TestCase):

    def test_Channel(self):

        c = OpenEXR.Channel(1)

    def test_File(self):

        # invalid argument
        with self.assertRaises(Exception):
            f = OpenEXR.File(1)

        # invalid number of arguments
        with self.assertRaises(Exception):
            f = OpenEXR.File("foo", "bar")

        # file not found
        filename = "/nonexistentfile.exr"
        with self.assertRaises(Exception):
            f = OpenEXR.File(filename)

        # file exists but is not an image
        filename = f"{test_dir}"
        with self.assertRaises(Exception):
            f = OpenEXR.File(filename)

        # Empty file object (useful it's possible to assign to it later)
        f = OpenEXR.File()
        self.assertEqual(f.filename, "")

        # no parts
        self.assertEqual(f.parts,[])
        with self.assertRaises(Exception):
            f.header()
        with self.assertRaises(Exception):
            f.channels()

        # 1-part file
        filename = f"{test_dir}/test.exr"
        f = OpenEXR.File(filename)
        self.assertEqual(f.filename, filename)

        # filename must be a string
        with self.assertRaises(Exception):
            f.filename = 1

        # invalid part
        with self.assertRaises(Exception):
            f.header(-1)
        with self.assertRaises(Exception):
            f.header(1)
        with self.assertRaises(Exception):
            f.channels(-1)
        with self.assertRaises(Exception):
            f.channels(1)

        self.assertEqual(len(f.parts),1)

        # invalid list (should be list of parts)
        with self.assertRaises(Exception):
            f = OpenEXR.File([1,2])
        with self.assertRaises(Exception):
            f = OpenEXR.File([OpenEXR.Part(),2])

        # empty header, empty channels
        f = OpenEXR.File({}, {})

        # bad header dict, bad channels dict
        with self.assertRaises(Exception):
            f = OpenEXR.File({1:1}, {})
        with self.assertRaises(Exception):
            f = OpenEXR.File({}, {1:1}) 
        with self.assertRaises(Exception):
            f = OpenEXR.File({}, {"A":1}) # bad value, shou

    def test_Part(self):

        with self.assertRaises(Exception):
            p = OpenEXR.Part(1)

        # bad header dict, bad channels dict
        with self.assertRaises(Exception):
            p = OpenEXR.Part({1:1}, {}, "party")
        with self.assertRaises(Exception):
            p = OpenEXR.Part({}, {1:1}, "party") 
        with self.assertRaises(Exception):
            p = OpenEXR.Part({}, {"A":1}, "party") # bad value, shou

        # test default type, compression
        p = OpenEXR.Part({}, {}, name="party")
        self.assertEqual(p.name(), "party")
        self.assertEqual(p.type(), OpenEXR.scanlineimage)
        self.assertEqual(p.compression(), OpenEXR.ZIP_COMPRESSION)
        self.assertEqual(p.width(), 0)
        self.assertEqual(p.height(), 0)
        self.assertEqual(p.channels, {})

        # test non-default type, compression
        p = OpenEXR.Part({"type" : OpenEXR.tiledimage,
                          "compression" : OpenEXR.NO_COMPRESSION}, {}, "party")
        self.assertEqual(p.type(), OpenEXR.tiledimage)
        self.assertEqual(p.compression(), OpenEXR.NO_COMPRESSION)

    def test_Channel(self):

        with self.assertRaises(Exception):
            OpenEXR.Channel(1)

        with self.assertRaises(Exception):
            OpenEXR.Channel("C", 2)

        C = OpenEXR.Channel("C", 2, 3)
        assert C.xSampling == 2
        assert C.ySampling == 3

        # not a 2D array
        with self.assertRaises(Exception):
            OpenEXR.Channel(np.array([0,0,0,0], dtype='uint32'))

        with self.assertRaises(Exception):
            OpenEXR.Channel("C", np.array([0,0,0,0], dtype='uint32'))
        with self.assertRaises(Exception):
            OpenEXR.Channel(np.array([0,0,0,0], dtype='uint32'), 2, 3)
        with self.assertRaises(Exception):
            OpenEXR.Channel("C", np.array([0,0,0,0], dtype='uint32'), 2, 3)

        # 2D array of unrecognized type
        width = 2
        height = 2
        with self.assertRaises(Exception):
            OpenEXR.Channel(np.array([0,0,0,0], dtype='uint8').reshape((height, width)))
        with self.assertRaises(Exception):
            OpenEXR.Channel("C", np.array([0,0,0,0], dtype='uint8').reshape((height, width)))
        with self.assertRaises(Exception):
            OpenEXR.Channel(np.array([0,0,0,0], dtype='uint8').reshape((height, width)), 2, 2)
        with self.assertRaises(Exception):
            OpenEXR.Channel("C", np.array([0,0,0,0], dtype='uint8').reshape((height, width)), 2, 2)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) Contributors to the OpenEXR Project.

import sys, os, tempfile, atexit
from do_run import do_run

print(f"testing exrmanifest: {sys.argv}")

src_dir = os.path.dirname (sys.argv[0])
exrmanifest = sys.argv[1]
exrinfo = sys.argv[2]
image_dir = sys.argv[3]
version = sys.argv[4]

test_images = {}
test_images["11"] = f"{image_dir}/11.deep.exr"
test_images["42"] = f"{image_dir}/42.deep.exr"
test_images["64"] = f"{image_dir}/64.deep.exr"
test_images["multivariate"] = f"{image_dir}/multivariate.deep.exr"
test_images["objectid"] = f"{image_dir}/objectid.deep.exr"

# no args = usage message, error
result = do_run ([exrmanifest], True)
assert result.stderr.startswith ("Usage: ")

# -h = usage message
result = do_run ([exrmanifest, "-h"])
assert result.stdout.startswith ("Usage: ")

result = do_run ([exrmanifest, "--help"])
assert result.stdout.startswith ("Usage: ")

# --version
result = do_run ([exrmanifest, "--version"])
assert result.stdout.startswith ("exrmanifest")
assert version in result.stdout

# invalid arguments
result = do_run ([exrmanifest, "foo.exr", "bar.exr"], True)

for image in test_images:
    result = do_run ([exrmanifest, test_images[image]])
    stdout_is = result.stdout
    with open (test_images[image] + ".txt", 'r') as file:
      stdout_should_be = "".join(l for l in file if not l.lstrip().startswith("#"))
      assert stdout_is == stdout_should_be

print("success")












#!/usr/bin/env python3

#
# SPDX-License-Identifier: BSD-3-Clause
# Copyright Contributors to the OpenEXR Project.
#

from __future__ import print_function
import sys
import os
import tempfile
import atexit
import unittest
import numpy as np
import fractions

import OpenEXR

test_dir = os.path.dirname(__file__)

outfilenames = []
def mktemp_outfilename():
    fd, outfilename = tempfile.mkstemp(".exr")
    os.close(fd)
    global outfilenames
    outfilenames += outfilename
    return outfilename

def cleanup():
    for outfilename in outfilenames:
        if os.path.isfile(outfilename):
            print(f"deleting {outfilename}")
            os.unlink(outfilename)
atexit.register(cleanup)


def equalWithRelError (x1, x2, e):
    return ((x1 - x2) if (x1 > x2) else (x2 - x1)) <= e * (x1 if (x1 > 0) else -x1)

def compare_files(lhs, rhs):

    if len(lhs.parts) != len(rhs.parts):
        raise Exception(f"#parts differs: {len(lhs.parts)} {len(rhs.parts)}")

    for Plhs, Prhs in zip(lhs.parts,rhs.parts):
        compare_parts(Plhs, Prhs)

def is_default(name, value):

    if name == "screenWindowWidth":
        return value == 1.0

    if name == "type":
        return value == OpenEXR.scanlineimage

    return True

def compare_parts(lhs, rhs):

    attributes = set(lhs.header.keys()).union(set(rhs.header.keys()))

    for a in attributes:
        if a in ["channels"]:
            continue

        if a not in lhs.header:
            if not is_default(a, rhs.header[a]):
                raise Exception(f"attribute {a} not in lhs header")
        elif a not in rhs.header:
            if not is_default(a, lhs.header[a]):
                raise Exception(f"attribute {a} not in rhs header")
        else:
            compare_attributes(a, lhs.header[a], rhs.header[a])

    if len(lhs.channels) != len(rhs.channels):
        raise Exception(f"#channels in {lhs.name} differs: {len(lhs.channels)} {len(rhs.channels)}")

    for c in lhs.channels.keys():
        compare_channels(lhs.channels[c], rhs.channels[c])

def compare_attributes(name, lhs, rhs):

    # convert tuples to array for comparison
    
    if isinstance(lhs, tuple):
        lhs = np.array(lhs)
        
    if isinstance(rhs, tuple):
        rhs = np.array(rhs)
        
    if isinstance(lhs, np.ndarray) and isinstance(rhs, np.ndarray):
        if lhs.shape != rhs.shape:
            raise Exception(f"attribute {name}: array shapes differ: {lhs} {rhs}")
        close = np.isclose(lhs, rhs, 1e-5)
        if not np.all(close):
            raise Exception(f"attribute {name}: arrays differ: {lhs} {rhs}")
    elif isinstance(lhs, float) and isinstance(rhs, float):
        if not equalWithRelError(lhs, rhs, 1e05):
            if math.isfinite(lhs) and math.isfinite(rhs):
                raise Exception(f"attribute {name}: floats differ: {lhs} {rhs}")
    elif lhs != rhs:
        raise Exception(f"attribute {name}: values differ: {lhs} {rhs}")

def compare_channels(lhs, rhs):

    if (lhs.name != rhs.name or
        lhs.type() != rhs.type() or
        lhs.xSampling != rhs.xSampling or
        lhs.ySampling != rhs.ySampling):
        raise Exception(f"channel {lhs.name} differs: {lhs.__repr__()} {rhs.__repr__()}")
    if lhs.pixels.shape != rhs.pixels.shape:
        raise Exception(f"channel {lhs.name}: image size differs: {lhs.pixels.shape} vs. {rhs.pixels.shape}")
        
    close = np.isclose(lhs.pixels, rhs.pixels, 1e-5)
    if not np.all(close):
        for i in np.argwhere(close==False):
            y,x = i
            if math.isfinite(lhs.pixels[y,x]) and math.isfinite(rhs.pixels[y,x]):
                raise Exception(f"channel {lhs.name}: pixels {i} differ: {lhs.pixels[y,x]} {rhs.pixels[y,x]}")


def print_file(f, print_pixels = False):

    print(f"file {f.filename}")
    print(f"parts:")
    parts = f.parts
    for p in parts:
        print(f"  part: {p.name()} {p.type()} {p.compression()} height={p.height()} width={p.width()}")
        h = p.header
        for a in h:
            print(f"    header[{a}] {h[a]}")
        for n,c in p.channels.items():
            print(f"    channel[{c.name}] shape={c.pixels.shape} strides={c.pixels.strides} {c.type()} {c.pixels.dtype}")
            if print_pixels:
                for y in range(0,c.pixels.shape[0]):
                    s = f"      {c.name}[{y}]:"
                    for x in range(0,c.pixels.shape[1]):
                        s += f" {c.pixels[y][x]}"
                    print(s)

def preview_pixels_equal(a, b):

    if a.shape != b.shape:
        return False

    for y in range(0,a.shape[0]):
        for x in range(0,a.shape[1]):
            if len(a[y][x]) != len(b[y][x]):
                for i in range(0,len(a[y][x])):
                    if a[y][x][i] != b[y][x][i]:
                        return False

    return True

class TestUnittest(unittest.TestCase):

    def setUp(self):
        # Print the name of the current test method
        print(f"Running test {self.id().split('.')[-1]}")

    def test_tuple(self):

        width = 5
        height = 10
        size = width * height
        Z = np.array([i for i in range(0,size)], dtype='uint32').reshape((height, width))
        channels = { "Z" : Z }

        header = {}
        header["t2i"] = (0,1)
        header["t2f"] = (2.3,4.5)
        header["t3i"] = (0,1,2)
        header["t3f"] = (3.4,5.6,7.8)

        header["a2f"] = np.array((1.2, 3.4), 'float32')
        header["a2d"] = np.array((1.2, 3.4), 'float64')
        header["a3f"] = np.array((1.2, 3.4, 5.6), 'float32')
        header["a3d"] = np.array((1.2, 3.4, 5.6), 'float64')

        header["a33f"] = np.identity(3, 'float32') 
        header["a33d"] = np.identity(3, 'float64') 
        header["a44f"] = np.identity(4, 'float32') 
        header["a44d"] = np.identity(4, 'float64') 

        outfilename = mktemp_outfilename()
        with OpenEXR.File(header, channels) as outfile:

            outfile.write(outfilename)

            with OpenEXR.File(outfilename) as infile:
                compare_files (infile, outfile)
                
        with self.assertRaises(Exception):
            header["v"] = (0,"x")
            with OpenEXR.File(header, channels) as outfile:
                outfile.write(outfilename)

        # tuple must be either all int or all float
        with self.assertRaises(Exception):
            header["v"] = (1,2.3)
            with OpenEXR.File(header, channels) as outfile:
                outfile.write(outfilename)

        os.remove(outfilename)

    def test_read_write(self):

        #
        # Read a file and write it back out, then read the freshly-written
        # file to validate it's the same.
        #

        infilename = f"{test_dir}/test.exr"
        with OpenEXR.File(infilename) as infile:

            outfilename = mktemp_outfilename()
            infile.write(outfilename)

            with OpenEXR.File(outfilename) as outfile:
                compare_files(outfile, infile)

    def test_keycode(self):

        filmMfcCode = 1
        filmType = 2
        prefix = 3
        count = 4
        perfOffset = 5
        perfsPerFrame = 6
        perfsPerCount = 20

        k = OpenEXR.KeyCode(filmMfcCode, filmType, prefix, count, perfOffset, perfsPerFrame, perfsPerCount)

        assert (k.filmMfcCode == filmMfcCode and
                k.filmType == filmType and
                k.prefix == prefix and
                k.count == count and
                k.perfOffset == perfOffset and
                k.perfsPerFrame == perfsPerFrame and
                k.perfsPerCount == perfsPerCount)

    def test_empty_header(self):

        # Construct a file from scratch and write it.

        width = 10
        height = 20
        size = width * height
        Z = np.array([i for i in range(0,size)], dtype='uint32').reshape((height, width))
        channels = { "Z" : OpenEXR.Channel(Z, 1, 1) }

        header = {}

        with OpenEXR.File(header, channels) as outfile:

            outfilename = mktemp_outfilename()
            outfile.write(outfilename)

            with OpenEXR.File(outfilename) as infile:
                compare_files (infile, outfile)

        os.remove(outfilename)
                
    def test_write_uint(self):

        # Construct a file from scratch and write it.

        width = 5
        height = 10
        size = width * height
        R = np.array([i for i in range(0,size)], dtype='uint32').reshape((height, width))
        G = np.array([i*2 for i in range(0,size)], dtype='uint32').reshape((height, width))
        B = np.array([i*3 for i in range(0,size)], dtype='uint32').reshape((height, width))
        A = np.array([i*5 for i in range(0,size)], dtype='uint32').reshape((height, width))
        channels = {
            "R" : OpenEXR.Channel(R, 1, 1),
            "G" : OpenEXR.Channel(G, 1, 1),
            "B" : OpenEXR.Channel(B, 1, 1),
            "A" : OpenEXR.Channel(A, 1, 1), 
        }

        header = {}

        with OpenEXR.File(header, channels) as outfile:

            # confirm that the write assigned names to the channels
            self.assertEqual(outfile.channels()['A'].name, "A")

            outfilename = mktemp_outfilename()
            outfile.write(outfilename)

            # Verify reading it back gives the same data
            with OpenEXR.File(outfilename, separate_channels=True) as infile:

                compare_files(infile, outfile)
        
        os.remove(outfilename)

    def test_write_half(self):

        # Construct a file from scratch and write it.

        width = 10
        height = 20
        size = width * height
        R = np.array([i for i in range(0,size)], dtype='e').reshape((height, width))
        G = np.array([i*10 for i in range(0,size)], dtype='e').reshape((height, width))
        B = np.array([i*100 for i in range(0,size)], dtype='e').reshape((height, width))
        A = np.array([i/size for i in range(0,size)], dtype='e').reshape((height, width))
        channels = {
            "A" : OpenEXR.Channel("A", A, 1, 1), 
            "B" : OpenEXR.Channel("B", B, 1, 1),
            "G" : OpenEXR.Channel("G", G, 1, 1),
            "R" : OpenEXR.Channel("R", R, 1, 1)
        }

        header = {}

        with OpenEXR.File(header, channels) as outfile:

            outfilename = mktemp_outfilename()
            outfile.write(outfilename)

            # Verify reading it back gives the same data
            with OpenEXR.File(outfilename, separate_channels=True) as infile:
                compare_files (infile, outfile)
                if "name" in infile.header():
                   raise Exception(f"name attribute was added to single part half file")

        os.remove(outfilename)

    def test_write_tiles(self):

        # Construct a file from scratch and write it.

        width = 10
        height = 20
        size = width * height
        R = np.array([i for i in range(0,size)], dtype='e').reshape((height, width))
        G = np.array([i*10 for i in range(0,size)], dtype='e').reshape((height, width))
        B = np.array([i*100 for i in range(0,size)], dtype='e').reshape((height, width))
        A = np.array([i/size for i in range(0,size)], dtype='e').reshape((height, width))
        channels = {
            "A" : OpenEXR.Channel("A", A, 1, 1), 
            "B" : OpenEXR.Channel("B", B, 1, 1),
            "G" : OpenEXR.Channel("G", G, 1, 1),
            "R" : OpenEXR.Channel("R", R, 1, 1)
        }

        header = { "type" : OpenEXR.tiledimage,
                   "tiles" : OpenEXR.TileDescription() }

        with OpenEXR.File(header, channels) as outfile:
            outfilename = mktemp_outfilename()
            outfile.write(outfilename)

            # Verify reading it back gives the same data
            with OpenEXR.File(outfilename, separate_channels=True) as infile:
                compare_files(infile, outfile)

        os.remove(outfilename)

    def test_modify_in_place(self):

        #
        # Test modifying header attributes in place
        #

        infilename = f"{test_dir}/test.exr"
        with OpenEXR.File(infilename, separate_channels=True) as f:

            # set the value of an existing attribute
            par = 2.3
            f.parts[0].header["pixelAspectRatio"] = par

            # add a new attribute
            f.parts[0].header["foo"] = "bar"

            dt = np.dtype({
                "names": ["r", "g", "b", "a"],
                "formats": ["u4", "u4", "u4", "u4"],
                "offsets": [0, 4, 8, 12],
            })
            pwidth = 3
            pheight = 3
            psize = pwidth * pheight
            P = np.array([ [(0,0,0,0), (1,1,1,1), (2,2,2,2) ],
                           [(3,3,3,3), (4,4,4,4), (5,5,5,5) ],
                           [(6,6,6,6), (7,7,7,7), (8,8,8,8) ] ], dtype=dt).reshape((pwidth,pheight))
            f.parts[0].header["preview"] = OpenEXR.PreviewImage(P)

            # Modify a pixel value
            f.parts[0].channels["R"].pixels[0][1] = 42.0
            f.channels()["G"].pixels[2][3] = 666.0

            # write to a new file
            outfilename = mktemp_outfilename()
            f.write(outfilename)

        # read the new file
        with OpenEXR.File(outfilename, separate_channels=True) as m:

            # validate the values are the same
            eps = 1e-5
            mpar = m.parts[0].header["pixelAspectRatio"]
            assert equalWithRelError(m.parts[0].header["pixelAspectRatio"], par, eps)
            assert m.parts[0].header["foo"] == "bar"

            assert preview_pixels_equal(m.parts[0].header["preview"].pixels, P)

            assert equalWithRelError(m.parts[0].channels["R"].pixels[0][1], 42.0, eps)
            assert equalWithRelError(m.parts[0].channels["G"].pixels[2][3], 666.0, eps)

    def test_header_only(self):

        #
        # Test modifying header when reading just the header.
        #

        multipartfilename = f"{test_dir}/multipart.exr"

        height, width = (20, 10)
        Z = np.zeros((height, width), dtype='f')
        P0 = OpenEXR.Part({"compression" : OpenEXR.ZIP_COMPRESSION,
                           "type" : OpenEXR.scanlineimage,
                           "foo" : "foo0",
                           "n" : 9,
                           "d" : 1.0 },
                          {"Z" : Z })

        R = np.random.rand(height, width).astype('f')
        G = np.random.rand(height, width).astype('f')
        B = np.random.rand(height, width).astype('f')
        P1 = OpenEXR.Part({"compression" : OpenEXR.ZIP_COMPRESSION,
                           "type" : OpenEXR.scanlineimage,
                           "foo" : "foo1",
                           "m" : 9,
                           "D" : 1.0 },
                          { "R" : R, "G" : G, "B" : B })

        f = OpenEXR.File([P0, P1])
        f.write(multipartfilename)

        # validate that the file has the proper values
        with OpenEXR.File(multipartfilename) as o:
            assert o.parts[0].name() == "Part0"
            assert o.parts[0].width() == 10
            assert o.parts[0].height() == 20
            assert o.parts[0].header["foo"] == "foo0"
            assert o.parts[0].header["n"] == 9
            assert o.parts[1].name() == "Part1"
            assert o.parts[1].width() == 10
            assert o.parts[1].height() == 20
            assert o.parts[1].header["foo"] == "foo1"
            assert o.parts[1].header["m"] == 9

        with OpenEXR.File(multipartfilename, separate_channels=True, header_only=True) as f:

            # change the value of an existing attribute
            f.parts[0].header["foo"] = "foo0!"
            f.parts[0].header["n"] = 99
            f.parts[1].header["foo"] = "foo1!"
            f.parts[1].header["m"] = 42

            # add a new attribute
            f.parts[0].header["bar"] = "bar0"
            f.parts[1].header["bar"] = "bar1"
            
            del f.parts[0].header["d"]
            del f.parts[1].header["D"]
            
            # write to a new file
            outfilename = mktemp_outfilename()
            f.write(outfilename)

        with OpenEXR.File(multipartfilename, separate_channels=True) as f:

            # read the new file
            with OpenEXR.File(outfilename, separate_channels=True) as m:

                # f has the original values
                # m has the modified values
                # Modify f again manually and validate that they match
            
                f.parts[0].header["foo"] = "foo0!"
                f.parts[0].header["n"] = 99
                f.parts[1].header["foo"] = "foo1!"
                f.parts[1].header["m"] = 42
                f.parts[0].header["bar"] = "bar0"
                f.parts[1].header["bar"] = "bar1"
                del f.parts[0].header["d"]
                del f.parts[1].header["D"]

                compare_files(f, m)

    def test_preview_image(self):

        width = 5
        height = 10
        size = width * height
        Z = np.array([i*5 for i in range(0,size)], dtype='uint32').reshape((height, width))
        channels = { "Z" : OpenEXR.Channel("Z", Z, 1, 1) }

        dt = np.dtype({
            "names": ["r", "g", "b", "a"],
            "formats": ["u4", "u4", "u4", "u4"],
            "offsets": [0, 4, 8, 12],
        })
        pwidth = 3
        pheight = 3
        psize = pwidth * pheight
        P = np.array([(i,i,i,i) for i in range(0,psize)], dtype=dt).reshape((pwidth,pheight))

        header = {}
        header["preview"] = OpenEXR.PreviewImage(P)

        with OpenEXR.File(header, channels) as outfile:

            outfilename = mktemp_outfilename()
            outfile.write(outfilename)

            with OpenEXR.File(outfilename) as infile:

                Q = infile.header()["preview"].pixels

                assert preview_pixels_equal(P, Q)

                compare_files (infile, outfile)

        os.remove(outfilename)

    def test_write_float(self):

        # Construct a file from scratch and write it.

        width = 50
        height = 1
        size = width * height
        R = np.array([i for i in range(0,size)], dtype='f').reshape((height, width))
        G = np.array([i*10 for i in range(0,size)], dtype='f').reshape((height, width))
        B = np.array([i*100 for i in range(0,size)], dtype='f').reshape((height, width))
        A = np.array([i*1000 for i in range(0,size)], dtype='f').reshape((height, width))
        channels = {
            "R" : OpenEXR.Channel("R", R, 1, 1),
            "G" : OpenEXR.Channel("G", G, 1, 1),
            "B" : OpenEXR.Channel("B", B, 1, 1),
            "A" : OpenEXR.Channel("A", A, 1, 1)
        }

        header = {}
        header["floatvector"] = [1.0, 2.0, 3.0]
        header["stringvector"] = ["do", "re", "me"]
        header["chromaticities"] = (1.0,2.0, 3.0,4.0, 5.0,6.0,7.0,8.0)
        header["box2i"] = ((0,1), (2,3))
        header["box2f"] = ((0.0,1.0), (2.0,3.0))
        header["bytes"] = OpenEXR.Bytes(b"\x76\x2f\x31\x01", "guess to win a prize")
        header["bytes0"] = OpenEXR.Bytes(b'')
        header["opaque"] = OpenEXR.OpaqueAttribute(b'',"Hello")
        header["compression"] = OpenEXR.ZIPS_COMPRESSION
        header["double"] = np.array([42000.0], 'float64')
        header["float"] = 4.2
        header["int"] = 42
        header["keycode"] = OpenEXR.KeyCode(0,0,0,0,0,4,64)
        header["lineorder"] = OpenEXR.INCREASING_Y
        header["m33f"] = np.identity(3, 'float32') 
        header["m33d"] = np.identity(3, 'float64') 
        header["m44f"] = np.identity(4, 'float32') 
        header["m44d"] = np.identity(4, 'float64') 
        header["rational"] = fractions.Fraction(1,3)
        header["string"] = "stringy"
        header["timecode"] = OpenEXR.TimeCode(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18)
        header["v2i"] = (1,2)
        header["v2f"] = (1.2,3.4)
        header["v2d"] = (1.2,3.4)
        header["v3i"] = (1,2,3)
        header["v3f"] = (1.2,3.4,5.6)
        header["v3d"] = (1.2,3.4,5.6)

        with OpenEXR.File(header, channels) as outfile:

            outfilename = mktemp_outfilename()
            outfile.write(outfilename)
                  
            # Verify reading it back gives the same data
            with OpenEXR.File(outfilename, separate_channels=True) as infile:
                compare_files (infile, outfile)

        os.remove(outfilename)

    def test_write_2part(self):

        #
        # Construct a 2-part file by replicating the header and channels
        #

        width = 10
        height = 20
        size = width * height
        R = np.array([i for i in range(0,size)], dtype='f').reshape((height, width))
        G = np.array([i*10 for i in range(0,size)], dtype='f').reshape((height, width))
        B = np.array([i*100 for i in range(0,size)], dtype='f').reshape((height, width))
        A = np.array([i*1000 for i in range(0,size)], dtype='f').reshape((height, width))
        channels = {
            "R" : OpenEXR.Channel("R", R, 1, 1),
            "G" : OpenEXR.Channel("G", G, 1, 1),
            "B" : OpenEXR.Channel("B", B, 1, 1),
            "A" : OpenEXR.Channel("A", A, 1, 1)
        }

        pwidth = 3
        pheight = 3
        psize = pwidth * pheight

        dt = np.dtype({
            "names": ["r", "g", "b", "a"],
            "formats": ["u4", "u4", "u4", "u4"],
            "offsets": [0, 4, 8, 12],
        })
        P = np.array([(i,i,i,i) for i in range(0,psize)], dtype=dt).reshape((pwidth,pheight))

        def make_header():
            header = {}
            header["floatvector"] = [1.0, 2.0, 3.0]
            return header
            header["stringvector"] = ["do", "re", "me"]
            header["chromaticities"] = (1.0,2.0, 3.0,4.0, 5.0,6.0,7.0,8.0)
            header["box2i"] = ((0,1),(2,3))
            header["box2f"] = ((0.0,1.0),(2.0,3.0))
            header["compression"] = OpenEXR.ZIPS_COMPRESSION
            header["double"] = np.array([42000.0], 'float64')
            header["float"] = 4.2
            header["int"] = 42
            header["keycode"] = OpenEXR.KeyCode(0,0,0,0,0,4,64)
            header["lineorder"] = OpenEXR.INCREASING_Y
            header["m33f"] = np.identity(3, 'float32')
            header["m33d"] = np.identity(3, 'float64')
            header["m44f"] = np.identity(4, 'float32')
            header["m44d"] = np.identity(4, 'float64')
            header["preview"] = OpenEXR.PreviewImage(P)
            header["rational"] = fractions.Fraction(1,3)
            header["string"] = "stringy"
            header["timecode"] = OpenEXR.TimeCode(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18)
            header["v2i"] = (1,2)
            header["v2f"] = (1.2,3.4)
            header["v2d"] = (1.2,3.4)
            header["v3i"] = (1,2,3)
            header["v3f"] = (1.2,3.4,5.6)
            header["v3d"] = (1.2,3.4,5.6)
            return header

        header1 = make_header()
        header2 = make_header()

        P1 = OpenEXR.Part(header1, channels, "Part1")
        P2 = OpenEXR.Part(header2, channels, "Part2")

        parts = [P1, P2]
        with OpenEXR.File(parts) as outfile2:

            outfilename = mktemp_outfilename()
            outfile2.write(outfilename)

            # Verify reading it back gives the same data
            with OpenEXR.File(outfilename, separate_channels=True) as i:
                compare_files (i, outfile2)

    def test_gil_released_during_io(self):

        #
        # Verify that the GIL is released during I/O operations by checking
        # that a background thread can make progress while OpenEXR performs
        # file operations.
        #

        import threading
        import time

        counter = [0]
        stop_flag = [False]

        def background_worker():
            while not stop_flag[0]:
                counter[0] += 1
                time.sleep(0.001)

        # Create a moderately large image to ensure I/O takes measurable time
        width = 1000
        height = 1000
        size = width * height
        R = np.random.rand(height, width).astype('f')
        G = np.random.rand(height, width).astype('f')
        B = np.random.rand(height, width).astype('f')
        channels = {
            "R": OpenEXR.Channel("R", R),
            "G": OpenEXR.Channel("G", G),
            "B": OpenEXR.Channel("B", B),
        }

        # Start background thread
        thread = threading.Thread(target=background_worker)
        thread.start()

        # Let thread run briefly to establish baseline
        time.sleep(0.05)
        count_before = counter[0]

        # Perform I/O - if GIL is released, background thread will increment counter
        outfilename = mktemp_outfilename()
        with OpenEXR.File({}, channels) as outfile:
            outfile.write(outfilename)

        count_after_write = counter[0]

        # Now test reading
        with OpenEXR.File(outfilename) as infile:
            _ = infile.channels()

        count_after_read = counter[0]

        # Stop background thread
        stop_flag[0] = True
        thread.join()

        # If GIL was released, background thread should have made progress during I/O
        write_progress = count_after_write - count_before
        read_progress = count_after_read - count_after_write

        self.assertGreater(write_progress, 0,
            f"Background thread made no progress during write (GIL not released?)")
        self.assertGreater(read_progress, 0,
            f"Background thread made no progress during read (GIL not released?)")

        os.remove(outfilename)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) Contributors to the OpenEXR Project.

import sys, os, tempfile, atexit
from do_run import do_run

print(f"testing exrmultiview: {' '.join(sys.argv)}")

exrmultiview = sys.argv[1]
exrinfo = sys.argv[2]
image_dir = sys.argv[3]
version = sys.argv[4]

left_image = f"{image_dir}/GammaChart.exr"
right_image = f"{image_dir}/GrayRampsHorizontal.exr"

result = do_run  ([exrmultiview], True)
assert "Usage:" in result.stderr

# -h = usage message
result = do_run  ([exrmultiview, "-h"])
assert result.stdout.startswith ("Usage: ")

result = do_run  ([exrmultiview, "--help"])
assert result.stdout.startswith ("Usage: ")

# --version
result = do_run  ([exrmultiview, "--version"])
assert result.stdout.startswith ("exrmultiview")
assert version in result.stdout

fd, outimage = tempfile.mkstemp(".exr")
os.close(fd)

def cleanup():
    print(f"deleting {outimage}")
    os.unlink(outimage)
atexit.register(cleanup)

result = do_run ([exrmultiview, "left", left_image, "right", right_image, outimage])

result = do_run  ([exrinfo, outimage])
print(" ".join(result.args))
assert result.returncode == 0
try:
    assert('\'B\': half samp 1 1' in result.stdout)
    assert('\'G\': half samp 1 1' in result.stdout)
    assert('\'R\': half samp 1 1' in result.stdout)
    assert('\'right.Y\': half samp 1 1' in result.stdout)
except AssertionError:
    print(result.stdout)
    raise

print("success")

